#
# ImageViewPg.py -- a backend for Ginga using javascript and
#      HTML5 canvas and websockets
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

from ginga import ImageView, Mixins, Bindings
from ginga.canvas.mixins import DrawingMixin, CanvasMixin, CompoundMixin
from ginga.canvas import render
from ginga.util.toolbox import ModeIndicator


default_html_fmt = 'jpeg'


class ImageViewPgError(ImageView.ImageViewError):
    pass


class ImageViewPg(ImageView.ImageViewBase):

    def __init__(self, logger=None, rgbmap=None, settings=None):
        ImageView.ImageViewBase.__init__(self, logger=logger,
                                         rgbmap=rgbmap,
                                         settings=settings)

        self.pgcanvas = None

        # format for rendering image on HTML5 canvas
        # NOTE: 'jpeg' has much better performance than 'png', but can show
        # some artifacts, especially noticeable with small text
        self.t_.set_defaults(html5_canvas_format=default_html_fmt,
                             renderer='cairo')

        self.rgb_order = 'RGBA'
        # this should already be so, but just in case...
        self.defer_redraw = True

        # these will be assigned in set_widget()
        self.timer_redraw = None
        self.timer_msg = None

        self.renderer = None
        # Pick a renderer that can work with us
        renderers = ['cairo', 'agg', 'pil', 'opencv']
        preferred = self.t_['renderer']
        if preferred in renderers:
            renderers.remove(preferred)
        self.possible_renderers = [preferred] + renderers
        self.choose_best_renderer()

    def set_widget(self, canvas_w):
        """Call this method with the widget that will be used
        for the display.
        """
        self.logger.debug("set widget canvas_w=%s" % canvas_w)
        self.pgcanvas = canvas_w

        app = canvas_w.get_app()
        self.timer_redraw = app.make_timer()
        self.timer_redraw.add_callback('expired',
                                       lambda t: self.delayed_redraw())
        self.timer_msg = app.make_timer()
        self.timer_msg.add_callback('expired',
                                    lambda t: self.clear_onscreen_message())

        wd, ht = canvas_w.get_size()
        self.configure_window(wd, ht)

    def get_widget(self):
        return self.pgcanvas

    def choose_renderer(self, name):
        klass = render.get_render_class(name)
        self.renderer = klass(self)

        if self.pgcanvas is not None:
            wd, ht = self.pgcanvas_w.get_size()
            self.configure_window(wd, ht)

    def choose_best_renderer(self):
        for name in self.possible_renderers:
            try:
                self.choose_renderer(name)
                self.logger.info("best renderer available is '{}'".format(name))
                return
            except Exception as e:
                continue

        raise ImageViewPgError("No valid renderers available: {}".format(str(self.possible_renderers)))

    def update_image(self):
        self.logger.debug("update_image pgcanvas=%s" % self.pgcanvas)
        if self.pgcanvas is None:
            return

        try:
            self.logger.debug("getting image as buffer...")
            format = self.t_.get('html5_canvas_format', default_html_fmt)

            buf = self.renderer.get_surface_as_rgb_format_bytes(
                format=format, quality=90)
            self.logger.debug("got '%s' RGB image buffer, len=%d" % (
                format, len(buf)))

            self.pgcanvas.do_update(buf)

        except Exception as e:
            self.logger.error("Couldn't update canvas: %s" % (str(e)))

    def reschedule_redraw(self, time_sec):
        if self.pgcanvas is not None:
            self.timer_redraw.stop()
            self.timer_redraw.start(time_sec)
        else:
            self.delayed_redraw()

    def get_plain_image_as_widget(self):
        """Used for generating thumbnails.  Does not include overlaid
        graphics.
        """
        image_buf = self.renderer.get_surface_as_rgb_format_buffer()
        return image_buf.getvalue()

    def save_plain_image_as_file(self, filepath, format='png', quality=90):
        """Used for generating thumbnails.  Does not include overlaid
        graphics.
        """
        pass

    def set_cursor(self, cursor):
        if self.pgcanvas is None:
            return
        #self.pgcanvas.config(cursor=cursor)

    def onscreen_message(self, text, delay=None, redraw=True):
        if self.pgcanvas is None:
            return
        self.timer_msg.stop()
        self.set_onscreen_message(text, redraw=redraw)
        if delay is not None:
            self.timer_msg.start(delay)

    def clear_onscreen_message(self):
        self.logger.debug("clearing message...")
        self.onscreen_message(None)

    def configure_window(self, width, height):
        self.configure_surface(width, height)

    def map_event(self, event):
        self.logger.info("window mapped to %dx%d" % (
            event.width, event.height))
        self.configure_window(event.width, event.height)
        self.redraw(whence=0)

    def resize_event(self, event):
        wd, ht = event.width, event.height
        # Not quite ready for prime-time--browser seems to mess with the
        # aspect ratio
        self.logger.info("canvas resized to %dx%d" % (wd, ht))
        self.configure_window(wd, ht)
        self.redraw(whence=0)

    def resize(self, width, height):
        """Resize our window to width x height.
        May not work---depending on how the HTML5 canvas is embedded.
        """
        # this shouldn't be needed
        self.configure_window(width, height)

        self.pgcanvas.resize(width, height)

        # hack to force a browser reload
        app = self.pgcanvas.get_app()
        app.do_operation('reload_page', id=self.pgcanvas.id)


class ImageViewEvent(ImageViewPg):

    def __init__(self, logger=None, rgbmap=None, settings=None):
        ImageViewPg.__init__(self, logger=logger, rgbmap=rgbmap,
                             settings=settings)

        self._button = 0

        # @$%&^(_)*&^ javascript!!
        # table mapping javascript key codes to ginga key names
        # see key_down_event() and key_up_event()
        #
        # https://www.cambiaresearch.com/articles/15/javascript-char-codes-key-codes
        self._keytbl = {
            8: 'backspace',
            9: 'tab',
            13: 'return',
            16: 'shift_l',
            #'shift_r': 'shift_r',
            17: 'control_l',
            #'control_r': 'control_r',
            18: 'alt_l',
            #'alt_r': 'alt_r',
            19: 'break',
            20: 'caps_lock',
            27: 'escape',
            32: 'space',
            33: 'page_up',
            34: 'page_down',
            35: 'end',
            36: 'home',
            37: 'left',
            38: 'up',
            39: 'right',
            40: 'down',
            45: 'insert',
            46: 'delete',
            65: 'a',
            66: 'b',
            67: 'c',
            68: 'd',
            69: 'e',
            70: 'f',
            71: 'g',
            72: 'h',
            73: 'i',
            74: 'j',
            75: 'k',
            76: 'l',
            77: 'm',
            78: 'n',
            79: 'o',
            80: 'p',
            81: 'q',
            82: 'r',
            83: 's',
            84: 't',
            85: 'u',
            86: 'v',
            87: 'w',
            88: 'x',
            89: 'y',
            90: 'z',
            91: 'super_l',
            92: 'super_r',
            93: 'menu_r',
            96: 'numpad_0',
            97: 'numpad_1',
            98: 'numpad_2',
            99: 'numpad_3',
            100: 'numpad_4',
            101: 'numpad_5',
            102: 'numpad_6',
            103: 'numpad_7',
            104: 'numpad_8',
            105: 'numpad_9',
            106: 'numpad_*',
            107: 'numpad_+',
            109: 'numpad_-',
            110: 'numpad_.',
            111: 'numpad_/',
            112: 'f1',
            113: 'f2',
            114: 'f3',
            115: 'f4',
            116: 'f5',
            117: 'f6',
            118: 'f7',
            119: 'f8',
            120: 'f9',
            121: 'f10',
            122: 'f11',
            123: 'f12',
            144: 'num_lock',
            145: 'scroll_lock',
            189: '-',
            186: ';',
            187: '=',
            188: ',',
            190: '.',
            191: '/',
            192: 'backquote',
            219: '[',
            220: 'backslash',
            221: ']',
            222: 'singlequote',
        }

        # this is an auxilliary table used to map shifted keys to names
        # see key_down_event() and key_up_event()
        self._keytbl2 = {
            ('shift_l', 'backquote'): '~',
            ('shift_l', '1'): '!',
            ('shift_l', '2'): '@',
            ('shift_l', '3'): '#',
            ('shift_l', '4'): '$',
            ('shift_l', '5'): '%',
            ('shift_l', '6'): '^',
            ('shift_l', '7'): '&',
            ('shift_l', '8'): '*',
            ('shift_l', '9'): '(',
            ('shift_l', '0'): ')',
            ('shift_l', 'a'): 'A',
            ('shift_l', 'b'): 'B',
            ('shift_l', 'c'): 'C',
            ('shift_l', 'd'): 'D',
            ('shift_l', 'e'): 'E',
            ('shift_l', 'f'): 'F',
            ('shift_l', 'g'): 'G',
            ('shift_l', 'h'): 'H',
            ('shift_l', 'i'): 'I',
            ('shift_l', 'j'): 'J',
            ('shift_l', 'k'): 'K',
            ('shift_l', 'l'): 'L',
            ('shift_l', 'm'): 'M',
            ('shift_l', 'n'): 'N',
            ('shift_l', 'o'): 'O',
            ('shift_l', 'p'): 'P',
            ('shift_l', 'q'): 'Q',
            ('shift_l', 'r'): 'R',
            ('shift_l', 's'): 'S',
            ('shift_l', 't'): 'T',
            ('shift_l', 'u'): 'U',
            ('shift_l', 'v'): 'V',
            ('shift_l', 'w'): 'W',
            ('shift_l', 'x'): 'X',
            ('shift_l', 'y'): 'Y',
            ('shift_l', 'z'): 'Z',
            ('shift_l', '-'): '_',
            ('shift_l', '='): '+',
            ('shift_l', '['): '{',
            ('shift_l', ']'): '}',
            ('shift_l', 'backslash'): '|',
            ('shift_l', ';'): ':',
            ('shift_l', 'singlequote'): 'doublequote',
            ('shift_l', ','): '<',
            ('shift_l', '.'): '>',
            ('shift_l', '/'): '?',
        }

        # this table is used to map special characters to character names
        # see key_press_event()
        self._keytbl3 = {
            '\\': 'backslash',
            '"': 'doublequote',
            "'": 'singlequote',
            "`": 'backquote',
            " ": 'space',
        }

        # list of keys for which javascript will give us a keydown event,
        # but not a keypress event.  We use this list to synthesize one.
        self._browser_problem_keys = set(['shift_l', 'control_l', 'alt_l',
                                          'super_l', 'super_r', 'menu_r',
                                          'escape', 'tab',
                                          'left', 'up', 'right', 'down',
                                          'insert', 'delete', 'home', 'end',
                                          'page_up', 'page_down',
                                          ])
        # Define cursors for pick and pan
        #hand = openHandCursor()
        hand = 'fleur'
        self.define_cursor('pan', hand)
        cross = 'cross'
        self.define_cursor('pick', cross)

        self._shifted = False

        for name in ('motion', 'button-press', 'button-release',
                     'key-press', 'key-release', 'drag-drop',
                     'scroll', 'map', 'focus', 'enter', 'leave',
                     'pinch', 'rotate', 'pan', 'swipe', 'tap'):
            self.enable_callback(name)

    def set_widget(self, canvas):
        super(ImageViewEvent, self).set_widget(canvas)

        # see event binding setup in Viewers.py

        #return self.make_callback('map')

    def transkey(self, keycode):
        self.logger.debug("key code in js '%d'" % (keycode))
        if keycode in self._keytbl:
            key = self._keytbl[keycode]
        else:
            key = chr(keycode)

        if self._shifted:
            try:
                key = self._keytbl2[('shift_l', key)]
            except KeyError:
                pass

        self.logger.debug("key name in ginga '%s'" % (key))
        return key

    def get_key_table(self):
        return self._keytbl

    def focus_event(self, event, has_focus):
        self.logger.debug("focus event: focus=%s" % (has_focus))
        return self.make_callback('focus', has_focus)

    def enter_notify_event(self, event):
        self.logger.debug("entering widget...")
        ## enter_focus = self.t_.get('enter_focus', False)
        ## if enter_focus:
        ##     self.pgcanvas.focus_set()
        return self.make_callback('enter')

    def leave_notify_event(self, event):
        self.logger.debug("leaving widget...")
        return self.make_callback('leave')

    def key_press_event(self, event):
        # For key_press_events, javascript reports the actual printable
        # key name.  We use a special keymap to just handle the few
        # characters for which we have special names
        keyname = event.key_name
        self.logger.debug("key press event, keyname=%s" % (keyname))
        if keyname in self._keytbl3:
            keyname = self._keytbl3[keyname]
        self.logger.debug("making key-press cb, key=%s" % (keyname))
        return self.make_ui_callback('key-press', keyname)

    def key_down_event(self, event):
        # For key down events, javascript only validly reports a key code.
        # We look up the code to determine the
        keycode = event.key_code
        self.logger.debug("key down event, keycode=%s" % (keycode))
        keyname = self.transkey(keycode)
        # special hack for modifiers
        if keyname == 'shift_l':
            self._shifted = True

        if keyname in self._browser_problem_keys:
            # JS doesn't report key press callbacks for certain keys
            # so we synthesize one here for those
            self.logger.debug("making key-press cb, key=%s" % (keyname))
            return self.make_ui_callback('key-press', keyname)
        return False

    def key_up_event(self, event):
        keycode = event.key_code
        self.logger.debug("key release event, keycode=%s" % (keycode))
        keyname = self.transkey(keycode)
        # special hack for modifiers
        if keyname == 'shift_l':
            self._shifted = False

        self.logger.debug("making key-release cb, key=%s" % (keyname))
        return self.make_ui_callback('key-release', keyname)

    def button_press_event(self, event):
        x = event.x
        y = event.y
        self.last_win_x, self.last_win_y = x, y
        button = 0
        button |= 0x1 << event.button
        self._button = button
        self.logger.debug("button event at %dx%d, button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()
        return self.make_ui_callback('button-press', button, data_x, data_y)

    def button_release_event(self, event):
        # event.button, event.x, event.y
        x = event.x
        y = event.y
        self.last_win_x, self.last_win_y = x, y
        button = 0
        button |= 0x1 << event.button
        self._button = 0
        self.logger.debug("button release at %dx%d button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()
        return self.make_ui_callback('button-release', button, data_x, data_y)

    def motion_notify_event(self, event):
        #button = 0
        button = self._button
        x, y = event.x, event.y
        self.last_win_x, self.last_win_y = x, y

        self.logger.debug("motion event at %dx%d, button=%x" % (x, y, button))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback('motion', button, data_x, data_y)

    def scroll_event(self, event):
        x, y = event.x, event.y
        delta = event.delta
        dx, dy = event.dx, event.dy
        self.last_win_x, self.last_win_y = x, y

        if (dx != 0 or dy != 0):
            # <= This browser gives us deltas for x and y
            # Synthesize this as a pan gesture event
            self.make_ui_callback('pan', 'start', 0, 0)
            self.make_ui_callback('pan', 'move', dx, dy)
            return self.make_ui_callback('pan', 'stop', 0, 0)

        # 15 deg is standard 1-click turn for a wheel mouse
        # delta usually returns +/- 1.0
        num_degrees = abs(delta) * 15.0

        direction = 0.0
        if delta > 0:
            direction = 0.0
        elif delta < 0:
            direction = 180.0
        self.logger.debug("scroll deg=%f direction=%f" % (
            num_degrees, direction))

        data_x, data_y = self.check_cursor_location()

        return self.make_ui_callback('scroll', direction, num_degrees,
                                     data_x, data_y)

    def drop_event(self, event):
        data = event.delta
        self.logger.debug("data=%s" % (str(data)))
        paths = data.split('\n')
        self.logger.debug("dropped text(s): %s" % (str(paths)))
        return self.make_ui_callback('drag-drop', paths)

    def pinch_event(self, event):
        self.logger.debug("pinch: event=%s" % (str(event)))
        state = 'move'
        if event.type == 'pinchstart' or event.isfirst:
            state = 'start'
        elif event.type == 'pinchend' or event.isfinal:
            state = 'end'
        rot = event.theta
        scale = event.scale
        self.logger.debug("pinch gesture rot=%f scale=%f state=%s" % (
            rot, scale, state))

        return self.make_ui_callback('pinch', state, rot, scale)

    def rotate_event(self, event):
        state = 'move'
        if event.type == 'rotatestart' or event.isfirst:
            state = 'start'
        elif event.type == 'rotateend' or event.isfinal:
            state = 'end'
        rot = event.theta
        self.logger.debug("rotate gesture rot=%f state=%s" % (
            rot, state))

        return self.make_ui_callback('rotate', state, rot)

    def pan_event(self, event):
        state = 'move'
        if event.type == 'panstart' or event.isfirst:
            state = 'start'
        elif event.type == 'panend' or event.isfinal:
            state = 'end'
        # TODO: need to know which ones to flip
        dx, dy = -event.dx, event.dy
        self.logger.debug("pan gesture dx=%f dy=%f state=%s" % (
            dx, dy, state))

        return self.make_ui_callback('pan', state, dx, dy)

    def swipe_event(self, event):
        if event.isfinal:
            state = 'end'  # noqa
            self.logger.debug("swipe gesture event=%s" % (str(event)))
            ## self.logger.debug("swipe gesture hdir=%s vdir=%s" % (
            ##     hdir, vdir))
            ## return self.make_ui_callback('swipe', state, hdir, vdir)

    def tap_event(self, event):
        if event.isfinal:
            state = 'end'  # noqa
            self.logger.debug("tap gesture event=%s" % (str(event)))


class ImageViewZoom(Mixins.UIMixin, ImageViewEvent):

    # class variables for binding map and bindings can be set
    bindmapClass = Bindings.BindingMapper
    bindingsClass = Bindings.ImageViewBindings

    @classmethod
    def set_bindingsClass(cls, klass):
        cls.bindingsClass = klass

    @classmethod
    def set_bindmapClass(cls, klass):
        cls.bindmapClass = klass

    def __init__(self, logger=None, rgbmap=None, settings=None,
                 bindmap=None, bindings=None):
        ImageViewEvent.__init__(self, logger=logger, rgbmap=rgbmap,
                                settings=settings)
        Mixins.UIMixin.__init__(self)

        self.ui_set_active(True)

        if bindmap is None:
            bindmap = ImageViewZoom.bindmapClass(self.logger)
        self.bindmap = bindmap
        bindmap.register_for_events(self)

        if bindings is None:
            bindings = ImageViewZoom.bindingsClass(self.logger)
        self.set_bindings(bindings)

    def get_bindmap(self):
        return self.bindmap

    def get_bindings(self):
        return self.bindings

    def set_bindings(self, bindings):
        self.bindings = bindings
        bindings.set_bindings(self)

    def center_cursor(self):
        # NOP
        pass

    def position_cursor(self, data_x, data_y):
        # NOP
        pass


class CanvasView(ImageViewZoom):

    def __init__(self, logger=None, settings=None, rgbmap=None,
                 bindmap=None, bindings=None):
        ImageViewZoom.__init__(self, logger=logger, settings=settings,
                               rgbmap=rgbmap,
                               bindmap=bindmap, bindings=bindings)

        # Needed for UIMixin to propagate events correctly
        self.objects = [self.private_canvas]

    def set_canvas(self, canvas, private_canvas=None):
        super(CanvasView, self).set_canvas(canvas,
                                           private_canvas=private_canvas)

        self.objects[0] = self.private_canvas


class ImageViewCanvas(ImageViewZoom,
                      DrawingMixin, CanvasMixin, CompoundMixin):

    def __init__(self, logger=None, rgbmap=None, settings=None,
                 bindmap=None, bindings=None):
        ImageViewZoom.__init__(self, logger=logger,
                               rgbmap=rgbmap,
                               settings=settings,
                               bindmap=bindmap,
                               bindings=bindings)
        CompoundMixin.__init__(self)
        CanvasMixin.__init__(self)
        DrawingMixin.__init__(self)

        # we are both a viewer and a canvas
        self.set_canvas(self, private_canvas=self)

        self._mi = ModeIndicator(self)


#END



#
# CL.py -- OpenCL functions for
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import sys
import os

import numpy as np

import pyopencl as cl

module_home = os.path.split(sys.modules[__name__].__file__)[0]


class CL(object):

    def __init__(self, filename=None):
        # TODO: we may want to choose GPU resources only
        self.ctx = cl.create_some_context()
        self.queue = cl.CommandQueue(self.ctx)

        if filename is not None:
            path = os.path.join(module_home, filename)
            self.load_program(path)

    def load_program(self, filename):
        #read in the OpenCL source file as a string
        with open(filename, 'r') as in_f:
            buf = in_f.read()
        self.load_program_buf(buf)

    def load_program_buf(self, buf):
        #create the program
        self.program = cl.Program(self.ctx, buf).build()

    def rotate(self, data_np, theta, rotctr_x=None, rotctr_y=None,
               clip_val=0,
               out=None, out_wd=0, out_ht=0, out_dx=0, out_dy=0):

        sin_theta = np.sin(np.radians(theta))
        cos_theta = np.cos(np.radians(theta))
        height, width = data_np.shape[:2]

        if rotctr_x is None:
            rotctr_x, rotctr_y = width // 2, height // 2

        mf = cl.mem_flags

        # convert to float64
        dtype = data_np.dtype
        data_np = np.ascontiguousarray(data_np, dtype=np.float64)

        if out is None:
            # no output array specified
            if out_ht == 0:
                # no desired output size specified--use dimensions of input
                # a clip, basically
                out_ht, out_wd = data_np.shape[:2]
            out_shape = (out_ht, out_wd) + data_np.shape[2:]
            out = np.empty(out_shape, dtype=data_np.dtype)
        else:
            # get dimensions of output array
            out_ht, out_wd = out.shape[:2]

        assert out.shape[2:] == data_np.shape[2:], ValueError(">2D dimensions don't match")

        numbytes = out_ht * out_wd * np.float64(0).nbytes

        #create OpenCL buffers on devices
        data_np = np.ascontiguousarray(data_np)
        src_buf = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR,
                            hostbuf=data_np)
        dst_buf = cl.Buffer(self.ctx, mf.WRITE_ONLY, numbytes)

        self.program.image_rotate_float64(
            self.queue, [out_ht, out_wd], None, src_buf, dst_buf,
            np.int32(rotctr_x), np.int32(rotctr_y),
            np.int32(width), np.int32(height),
            np.int32(out_wd), np.int32(out_ht),
            np.int32(out_dx), np.int32(out_dy),
            np.float64(sin_theta), np.float64(cos_theta),
            np.float64(clip_val))

        if dtype == np.float64:
            out_np = out
        else:
            out_np = np.empty(out_shape, dtype=np.float64)

        cl.enqueue_read_buffer(self.queue, dst_buf, out_np).wait()
        #cl.enqueue_copy(self.queue, out_np, dst_buf).wait()

        res = out_np.astype(dtype, copy=False)
        if out is not None:
            out[...] = res
        else:
            out = res

        return out

    def rotate_clip(self, data_np, theta, rotctr_x=None, rotctr_y=None,
                    clip_val=0, out=None):

        sin_theta = np.sin(np.radians(theta))
        cos_theta = np.cos(np.radians(theta))
        height, width = data_np.shape[:2]

        if rotctr_x is None:
            rotctr_x, rotctr_y = width // 2, height // 2

        mf = cl.mem_flags

        # convert to float64
        dtype = data_np.dtype
        data_np = np.ascontiguousarray(data_np, dtype=np.float64)

        # clipped array is same size as original
        out_ht, out_wd = data_np.shape[:2]
        out_dx, out_dy = 0, 0
        numbytes = data_np.nbytes

        #create OpenCL buffers on devices
        data_np = np.ascontiguousarray(data_np)
        src_buf = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR,
                            hostbuf=data_np)
        dst_buf = cl.Buffer(self.ctx, mf.WRITE_ONLY, numbytes)

        self.program.image_rotate_float64(
            self.queue, [height, width], None, src_buf, dst_buf,
            np.int32(rotctr_x), np.int32(rotctr_y),
            np.int32(width), np.int32(height),
            np.int32(out_wd), np.int32(out_ht),
            np.int32(out_dx), np.int32(out_dy),
            np.float64(sin_theta), np.float64(cos_theta),
            np.float64(clip_val))

        out_np = np.empty_like(data_np)
        cl.enqueue_read_buffer(self.queue, dst_buf, out_np).wait()
        #cl.enqueue_copy(self.queue, out_np, dst_buf).wait()

        res = out_np.astype(dtype, copy=False)
        if out is not None:
            out[...] = res
        else:
            out = res

        return out

    def rotate_clip_uint32(self, data_np, theta, rotctr_x=None, rotctr_y=None,
                           clip_val=0, out=None):

        sin_theta = np.sin(np.radians(theta))
        cos_theta = np.cos(np.radians(theta))
        height, width = data_np.shape[:2]

        if rotctr_x is None:
            rotctr_x, rotctr_y = width // 2, height // 2

        mf = cl.mem_flags

        if out is None:
            out = np.empty_like(data_np)
        # clipped array is same size as original
        out_ht, out_wd = out.shape[:2]
        out_dx, out_dy = 0, 0

        #create OpenCL buffers on devices
        data_np = np.ascontiguousarray(data_np)
        src_buf = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR,
                            hostbuf=data_np)
        dst_buf = cl.Buffer(self.ctx, mf.WRITE_ONLY, out.nbytes)

        self.program.image_rotate_uint32(
            self.queue, [height, width], None, src_buf, dst_buf,
            np.int32(rotctr_x), np.int32(rotctr_y),
            np.int32(width), np.int32(height),
            np.int32(out_wd), np.int32(out_ht),
            np.int32(out_dx), np.int32(out_dy),
            np.float64(sin_theta), np.float64(cos_theta),
            np.uint32(clip_val))

        cl.enqueue_read_buffer(self.queue, dst_buf, out).wait()

        return out

    def rotate_uint32(self, data_np, theta, rotctr_x=None, rotctr_y=None,
                      clip_val=0,
                      out=None, out_wd=0, out_ht=0, out_dx=0, out_dy=0):

        sin_theta = np.sin(np.radians(theta))
        cos_theta = np.cos(np.radians(theta))
        height, width = data_np.shape[:2]

        if rotctr_x is None:
            rotctr_x, rotctr_y = width // 2, height // 2

        mf = cl.mem_flags

        if out is None:
            if out_ht == 0:
                out_ht, out_wd = data_np.shape[:2]
            out_shape = (out_ht, out_wd) + data_np.shape[2:]
            out = np.zeros(out_shape, dtype=data_np.dtype)
        else:
            out_ht, out_wd = out.shape[:2]

        assert out.shape[2:] == data_np.shape[2:], ValueError(">2D dimensions don't match")

        #create OpenCL buffers on devices
        data_np = np.ascontiguousarray(data_np)
        src_buf = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR,
                            hostbuf=data_np)
        dst_buf = cl.Buffer(self.ctx, mf.WRITE_ONLY, out.nbytes)

        self.program.image_rotate_uint32(
            self.queue, [out_ht, out_wd], None, src_buf, dst_buf,
            np.int32(rotctr_x), np.int32(rotctr_y),
            np.int32(width), np.int32(height),
            np.int32(out_wd), np.int32(out_ht),
            np.int32(out_dx), np.int32(out_dy),
            np.float64(sin_theta), np.float64(cos_theta),
            np.uint32(clip_val))

        cl.enqueue_read_buffer(self.queue, dst_buf, out).wait()

        return out

    def transform_uint32(self, data_np,
                         flip_x=False, flip_y=False, swap_xy=False,
                         out=None):

        height, width = data_np.shape[:2]

        new_ht, new_wd = height, width
        if swap_xy:
            new_ht, new_wd = width, height
        new_size = [new_ht, new_wd] + list(data_np.shape[2:])

        mf = cl.mem_flags

        #create OpenCL buffers on devices
        data_np = np.ascontiguousarray(data_np)
        src_buf = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR,
                            hostbuf=data_np)
        dst_buf = cl.Buffer(self.ctx, mf.WRITE_ONLY, data_np.nbytes)

        self.program.image_transform_uint32(
            self.queue, [height, width], None, src_buf, dst_buf,
            np.int32(width), np.int32(height),
            np.int32(flip_x), np.int32(flip_y),
            np.int32(swap_xy))

        if out is None:
            out = np.empty_like(data_np).reshape(new_size)
        cl.enqueue_read_buffer(self.queue, dst_buf, out).wait()

        return out

    def resize_uint32(self, data_np, scale_x, scale_y, out=None):

        height, width = data_np.shape[:2]

        new_ht = int(height * scale_y)
        new_wd = int(width * scale_x)
        new_shape = [new_ht, new_wd] + list(data_np.shape[2:])

        mf = cl.mem_flags

        #create OpenCL buffers on devices
        data_np = np.ascontiguousarray(data_np)
        src_buf = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR,
                            hostbuf=data_np)

        num_bytes = new_ht * new_wd * np.uint32(0).nbytes
        dst_buf = cl.Buffer(self.ctx, mf.WRITE_ONLY, num_bytes)

        self.program.image_resize_uint32(
            self.queue, [new_ht, new_wd], None, src_buf, dst_buf,
            np.int32(width), np.int32(new_wd),
            np.float64(scale_x), np.float64(scale_y))

        if out is None:
            out = np.empty(new_shape, dtype=data_np.dtype)
        cl.enqueue_read_buffer(self.queue, dst_buf, out).wait()

        return out

    def resize(self, data_np, scale_x, scale_y, out=None):

        height, width = data_np.shape[:2]

        new_ht = int(height * scale_y)
        new_wd = int(width * scale_x)
        new_shape = [new_ht, new_wd] + list(data_np.shape[2:])

        mf = cl.mem_flags

        #create OpenCL buffers on devices
        dtype = data_np.dtype
        data_np = np.ascontiguousarray(data_np, dtype=np.float64)

        src_buf = cl.Buffer(self.ctx, mf.READ_ONLY | mf.COPY_HOST_PTR,
                            hostbuf=data_np)

        num_bytes = new_ht * new_wd * np.float64(0).nbytes
        dst_buf = cl.Buffer(self.ctx, mf.WRITE_ONLY, num_bytes)

        self.program.image_resize_float64(
            self.queue, [new_ht, new_wd], None, src_buf, dst_buf,
            np.int32(width), np.int32(new_wd),
            np.float64(scale_x), np.float64(scale_y))

        if dtype == np.float64:
            out_np = out
        else:
            out_np = np.empty(new_shape, dtype=np.float64)

        cl.enqueue_read_buffer(self.queue, dst_buf, out_np).wait()

        res = out_np.astype(dtype, copy=False)
        if out is not None:
            out[...] = res
        else:
            out = res

        return out

    def get_scaled_cutout_basic(self, data_np, x1, y1, x2, y2, scale_x, scale_y,
                                out=None):

        if (data_np.dtype == np.uint32) or ((data_np.dtype == np.uint8) and
                                            (len(data_np.shape) == 3)):
            newdata = self.resize_uint32(data_np[y1:y2 + 1, x1:x2 + 1, ...],
                                         scale_x, scale_y, out=out)
        else:
            newdata = self.resize(data_np[y1:y2 + 1, x1:x2 + 1],
                                  scale_x, scale_y, out=out)

        old_ht, old_wd = data_np.shape[:2]
        ht, wd = newdata.shape[:2]
        scale_x, scale_y = float(wd) / old_wd, float(ht) / old_ht

        return newdata, (scale_x, scale_y)

#END



# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
"""
Display masks from file (non-interative mode) on an image.

**Plugin Type: Local**

``TVMask`` is a local plugin, which means it is associated with a
channel.  An instance can be opened for each channel.

**Usage**

This plugin allows non-interactive display of mask by reading in a FITS
file, where non-zero is assumed to be masked data.

To display different masks (e.g., some masked as green and some as pink, as
shown above):

1. Select green from the drop-down menu. Alternately, enter desired
   alpha value.
2. Using "Load Mask" button, load the relevant FITS file.
3. Repeat (1) but now select pink from the drop-down menu.
4. Repeat (2) but choose another FITS file.
5. To display a third mask as pink too, repeat (4) without changing the
   drop-down menu.

Selecting an entry (or multiple entries) from the table listing will
highlight the mask(s) on the image. The highlight uses a pre-defined color and
alpha (customizable below).

You can also highlight all the masks within a region both on the image
and the table listing by drawing a rectangle on the image
while this plugin is active.

Pressing the "Hide" button will hide the masks but does not clear the
plugin's memory; That is, when you press "Show", the same masks will
reappear on the same image. However, pressing "Forget" will clear the masks
both from display and memory; That is, you will need to reload your file(s) to
recreate the masks.

To redraw the same masks with different color or alpha, press "Forget"
and repeat the steps above, as necessary.

If images of very different pointings/dimensions are displayed in the same
channel, masks that belong to one image but fall outside another will not
appear in the latter.

To create a mask that this plugin can read, one can use results from
the ``Drawing`` plugin (press "Create Mask" after drawing and save the
mask using ``SaveImage``), in addition to creating a FITS
file by hand using ``astropy.io.fits``, etc.

Used together with ``TVMark``, you can overlay both point sources and
masked regions in Ginga.

"""
# STDLIB
import re
import os

# THIRD-PARTY
import numpy as np
from astropy.io import fits

# GINGA
from ginga import colors
from ginga.GingaPlugin import LocalPlugin
from ginga.gw import Widgets
from ginga.misc import Bunch
from ginga.util.dp import masktorgb

# Need this for API doc to build without warning
try:
    from ginga.gw.GwHelp import FileSelection
except ImportError:
    pass

__all__ = ['TVMask']


class TVMask(LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super(TVMask, self).__init__(fv, fitsimage)

        self.layertag = 'tvmask-canvas'
        self.masktag = None
        self.maskhltag = None

        self._color_options = self._short_color_list()

        # User preferences. Some are just default values and can also be
        # changed by GUI.
        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_TVMask')
        self.settings.add_defaults(maskcolor='green', maskalpha=0.5,
                                   hlcolor='white', hlalpha=1.0)
        self.settings.load(onError='silent')
        self.maskcolor = self.settings.get('maskcolor', 'green')
        self.maskalpha = self.settings.get('maskalpha', 0.5)
        self.hlcolor = self.settings.get('hlcolor', 'white')
        self.hlalpha = self.settings.get('hlalpha', 1.0)

        # Display coords info table
        self.treeview = None
        self.tree_dict = Bunch.caselessDict()
        self.columns = [('No.', 'ID'), ('Filename', 'MASKFILE')]

        # Store results
        self._seqno = 1
        self._maskobjs = []
        self._treepaths = []

        self.dc = self.fv.get_draw_classes()

        canvas = self.dc.DrawingCanvas()
        canvas.enable_draw(True)
        canvas.enable_edit(False)
        canvas.set_callback('draw-event', self.hl_canvas2table_box)
        #canvas.set_callback('cursor-down', self.hl_canvas2table)
        canvas.register_for_cursor_drawing(self.fitsimage)
        canvas.set_surface(self.fitsimage)
        canvas.set_drawtype('rectangle', color='green', linestyle='dash')
        self.canvas = canvas

        fv.add_callback('remove-image', lambda *args: self.redo())

        self.gui_up = False

    # If user complains about lack of choices (!!!), we can remove this.
    def _short_color_list(self):
        """Color list is too long. Discard variations with numbers."""
        return [c for c in colors.get_colors() if not re.search(r'\d', c)]

    def build_gui(self, container):
        vbox, sw, self.orientation = Widgets.get_oriented_box(container)

        captions = (('Color:', 'label', 'mask color', 'combobox'),
                    ('Alpha:', 'label', 'mask alpha', 'entry'))
        w, b = Widgets.build_info(captions)
        self.w.update(b)

        combobox = b.mask_color
        for name in self._color_options:
            combobox.append_text(name)
        b.mask_color.set_index(self._color_options.index(self.maskcolor))
        b.mask_color.add_callback('activated', self.set_maskcolor_cb)

        b.mask_alpha.set_tooltip('Mask alpha (transparency)')
        b.mask_alpha.set_text(str(self.maskalpha))
        b.mask_alpha.add_callback('activated', lambda w: self.set_maskalpha())

        container.add_widget(w, stretch=0)

        treeview = Widgets.TreeView(auto_expand=True,
                                    sortable=True,
                                    selection='multiple',
                                    use_alt_row_color=True)
        self.treeview = treeview
        treeview.setup_table(self.columns, 2, 'ID')
        treeview.add_callback('selected', self.hl_table2canvas)
        container.add_widget(treeview, stretch=1)

        captions = (('Load Mask', 'button'),
                    ('Show', 'button', 'Hide', 'button', 'Forget', 'button'))
        w, b = Widgets.build_info(captions)
        self.w.update(b)

        b.load_mask.set_tooltip('Load mask image')
        b.load_mask.add_callback('activated', lambda w: self.load_mask_cb())

        b.show.set_tooltip('Show masks')
        b.show.add_callback('activated', lambda w: self.redo())

        b.hide.set_tooltip('Hide masks')
        b.hide.add_callback('activated', lambda w: self.clear_mask())

        b.forget.set_tooltip('Forget masks')
        b.forget.add_callback('activated', lambda w: self.forget_masks())

        container.add_widget(w, stretch=0)

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(3)

        btn = Widgets.Button('Close')
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn, stretch=0)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)

        container.add_widget(btns, stretch=0)

        self.gui_up = True

        # Initialize mask file selection dialog
        self.mfilesel = FileSelection(self.fv.w.root.get_widget(),
                                      all_at_once=True)

        # Populate table
        self.redo()

    def redo(self):
        """Image or masks have changed. Clear and redraw."""
        if not self.gui_up:
            return

        self.clear_mask()

        image = self.fitsimage.get_image()
        if image is None:
            return

        n_obj = len(self._maskobjs)
        self.logger.debug('Displaying {0} masks'.format(n_obj))
        if n_obj == 0:
            return

        # Display info table
        self.recreate_toc()

        # Draw on canvas
        self.masktag = self.canvas.add(self.dc.CompoundObject(*self._maskobjs))
        self.fitsimage.redraw()  # Force immediate redraw

    def clear_mask(self):
        """Clear mask from image.
        This does not clear loaded masks from memory."""
        if self.masktag:
            try:
                self.canvas.delete_object_by_tag(self.masktag, redraw=False)
            except Exception:
                pass

        if self.maskhltag:
            try:
                self.canvas.delete_object_by_tag(self.maskhltag, redraw=False)
            except Exception:
                pass

        self.treeview.clear()  # Clear table too
        self.fitsimage.redraw()  # Force immediate redraw

    def forget_masks(self):
        """Forget all loaded coordinates."""
        self._seqno = 1
        self._maskobjs = []
        self._treepaths = []
        self.tree_dict = Bunch.caselessDict()
        self.redo()

    # TODO: Support more formats?
    def load_file(self, filename):
        """Load mask image.

        Results are appended to previously loaded masks.
        This can be used to load mask per color.

        """
        if not os.path.isfile(filename):
            return

        self.logger.info('Loading mask image from {0}'.format(filename))

        try:
            # 0=False, everything else True
            dat = fits.getdata(filename).astype(np.bool)
        except Exception as e:
            self.logger.error('{0}: {1}'.format(e.__class__.__name__, str(e)))
            return

        key = '{0},{1}'.format(self.maskcolor, self.maskalpha)

        if key in self.tree_dict:
            sub_dict = self.tree_dict[key]
        else:
            sub_dict = {}
            self.tree_dict[key] = sub_dict

        # Add to listing
        seqstr = '{0:04d}'.format(self._seqno)  # Prepend 0s for proper sort
        sub_dict[seqstr] = Bunch.Bunch(ID=seqstr,
                                       MASKFILE=os.path.basename(filename))
        self._treepaths.append((key, seqstr))
        self._seqno += 1

        # Create mask layer
        obj = self.dc.Image(0, 0, masktorgb(
            dat, color=self.maskcolor, alpha=self.maskalpha))
        self._maskobjs.append(obj)

        self.redo()

    def load_files(self, filenames):
        """Load mask images.

        Results are appended to previously loaded masks.
        This can be used to load mask per color.

        """
        for filename in filenames:
            self.load_file(filename)

    def load_mask_cb(self):
        """Activate file dialog to select mask image."""
        self.mfilesel.popup('Load mask image', self.load_files,
                            initialdir='.', filename='FITS files (*.fits)')

    def recreate_toc(self):
        self.logger.debug('Recreating table of contents...')
        self.treeview.set_tree(self.tree_dict)

    def _rgbtomask(self, obj):
        """Convert RGB arrays from mask canvas object back to boolean mask."""
        dat = obj.get_image().get_data()  # RGB arrays
        return dat.sum(axis=2).astype(np.bool)  # Convert to 2D mask

    def hl_table2canvas(self, w, res_dict):
        """Highlight mask on canvas when user click on table."""
        objlist = []

        # Remove existing highlight
        if self.maskhltag:
            try:
                self.canvas.delete_object_by_tag(self.maskhltag, redraw=False)
            except Exception:
                pass

        for sub_dict in res_dict.values():
            for seqno in sub_dict:
                mobj = self._maskobjs[int(seqno) - 1]
                dat = self._rgbtomask(mobj)
                obj = self.dc.Image(0, 0, masktorgb(
                    dat, color=self.hlcolor, alpha=self.hlalpha))
                objlist.append(obj)

        # Draw on canvas
        if len(objlist) > 0:
            self.maskhltag = self.canvas.add(self.dc.CompoundObject(*objlist))

        self.fitsimage.redraw()  # Force immediate redraw

    def hl_canvas2table_box(self, canvas, tag):
        """Highlight all masks inside user drawn box on table."""
        self.treeview.clear_selection()

        # Remove existing box
        cobj = canvas.get_object_by_tag(tag)
        if cobj.kind != 'rectangle':
            return
        canvas.delete_object_by_tag(tag, redraw=False)

        # Remove existing highlight
        if self.maskhltag:
            try:
                canvas.delete_object_by_tag(self.maskhltag, redraw=True)
            except Exception:
                pass

        # Nothing to do if no masks are displayed
        try:
            obj = canvas.get_object_by_tag(self.masktag)
        except Exception:
            return

        if obj.kind != 'compound':
            return

        # Nothing to do if table has no data
        if len(self._maskobjs) == 0:
            return

        # Find masks that intersect the rectangle
        for i, mobj in enumerate(self._maskobjs):
            # The actual mask
            mask1 = self._rgbtomask(mobj)

            # The selected area
            rgbimage = mobj.get_image()
            mask2 = rgbimage.get_shape_mask(cobj)

            # Highlight mask with intersect
            if np.any(mask1 & mask2):
                self._highlight_path(self._treepaths[i])

    # NOTE: This does not work anymore when left click is used to draw box.
    def hl_canvas2table(self, canvas, button, data_x, data_y):
        """Highlight mask on table when user click on canvas."""
        self.treeview.clear_selection()

        # Remove existing highlight
        if self.maskhltag:
            try:
                canvas.delete_object_by_tag(self.maskhltag, redraw=True)
            except Exception:
                pass

        # Nothing to do if no masks are displayed
        try:
            obj = canvas.get_object_by_tag(self.masktag)
        except Exception:
            return

        if obj.kind != 'compound':
            return

        # Nothing to do if table has no data
        if len(self._maskobjs) == 0:
            return

        for i, mobj in enumerate(self._maskobjs):
            mask1 = self._rgbtomask(mobj)

            # Highlight mask covering selected cursor position
            if mask1[int(data_y), int(data_x)]:
                self._highlight_path(self._treepaths[i])

    def _highlight_path(self, hlpath):
        """Highlight an entry in the table and associated mask."""
        self.logger.debug('Highlighting {0}'.format(hlpath))
        self.treeview.select_path(hlpath)

        # TODO: Does not work in Qt. This is known issue in Ginga.
        self.treeview.scroll_to_path(hlpath)

    def set_maskcolor_cb(self, w, index):
        """Set color of mask."""
        self.maskcolor = self._color_options[index]

    def set_maskalpha(self):
        """Set alpha (transparency) of mask."""
        try:
            a = float(self.w.mask_alpha.get_text())
        except ValueError:
            self.logger.error('Cannot set mask alpha')
            self.w.mask_alpha.set_text(str(self.maskalpha))
            return

        if a < 0 or a > 1:
            self.logger.error('Alpha must be between 0 and 1, inclusive')
            self.w.mask_alpha.set_text(str(self.maskalpha))
            return

        self.maskalpha = a

    def close(self):
        self.fv.stop_local_plugin(self.chname, str(self))
        return True

    def start(self):
        # insert canvas, if not already
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.get_object_by_tag(self.layertag)
        except KeyError:
            # Add drawing layer
            p_canvas.add(self.canvas, tag=self.layertag)

        self.resume()

    def pause(self):
        self.canvas.ui_set_active(False)

    def resume(self):
        # turn off any mode user may be in
        self.modes_off()

        self.canvas.ui_set_active(True)
        self.fv.show_status('Press "Help" for instructions')

    def stop(self):
        # remove canvas from image
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.delete_object_by_tag(self.layertag)
        except Exception:
            pass

        self.canvas.update_canvas(whence=0)  # Force redraw
        self.gui_up = False
        self.fv.show_status('')

    def __str__(self):
        """
        This method should be provided and should return the lower case
        name of the plugin.
        """
        return 'tvmask'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_TVMask', package='ginga')

# END



#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
"""
``PixTable`` provides a way to check or monitor the pixel values in
a region.

**Plugin Type: Local**

``PixTable`` is a local plugin, which means it is associated with a channel.
An instance can be opened for each channel.

**Basic Use**

In the most basic use, simply move the cursor around the channel
viewer; an array of pixel values will appear in the "Pixel Values"
display in the plugin UI.  The center value is highlighted, and this
corresponds to the value under the cursor.

You can choose a 3x3, 5x5, 7x7, or 9x9 grid from the left-most
combobox control.  It may help to adjust the "Font Size" control
to prevent having the array values cut off on the sides.  You can
also enlarge the plugin workspace to see more of the table.

.. note:: The order of the value table shown will not necessarily match to
          the channel viewer if the images is flipped, transposed, or rotated.

**Using Marks**

When you set and select a mark, the pixel values will be shown
surrounding the mark instead of the cursor.  There can be any number
of marks, and they are each noted with a numbered "X".  Simply change the
mark drop down control to select a different mark and see the values
around it.  The currently selected mark is shown with a different color
than the others.

The marks will stay in position even if a new image is loaded and
they will show the values for the new image.  In this way you can
monitor the area around a spot if the image is updating frequently.

If the "Pan to mark" checkbox is selected, then when you select a
different mark from the mark control, the channel viewer will pan to
that mark.  This can be useful to inspect the same spots in several
different images, especially when zoomed in tight to the image.

.. note:: If you change the mark control back to "None", then the pixel
          table will again update as you move the cursor around the viewer.

The "Caption" box can be used to set a text annotation that will be
appended to the mark label when the next mark is created.  This can be
used to label a feature in the image, for example.

**Deleting Marks**

To delete a mark, select it in the mark control and then press the
button marked "Delete".  To delete all the marks, press the button
marked "Delete All".

**Moving Marks**

When the "Move" radio button is checked, and a mark is selected, then
clicking or dragging anywhere in the image will move the mark to that
location and update the pixel table.   If no mark is currently selected
then a new one will be created and moved.

**Drawing Marks**

When the "Draw" radio button is checked, then clicking and dragging creates
a new mark. The longer the draw, the bigger radius of the "X".

**Editing Marks**

When the "Edit" radio button is checked after a mark has been selected then
you can drag the control points of the mark to increase the radius of the
arms of the X or you can drag the bounding box to move the mark. If the
editing control points are not shown, simply click on the center of a mark
to enable them.

**Special Keys**

In "Move" mode the following keys are active:
- "n" will place a new mark at the site of the cursor
- "m" will move the current mark (if any) to the site of the cursor
- "d" will delete the current mark (if any)
- "j" will select the previous mark (if any)
- "k" will select the next mark (if any)

**User Configuration**

"""
import numpy as np

from ginga.gw import Widgets, Viewers
from ginga import GingaPlugin, colors

__all__ = ['PixTable']


class PixTable(GingaPlugin.LocalPlugin):

    def __init__(self, fv, fitsimage):
        # superclass defines some variables for us, like logger
        super(PixTable, self).__init__(fv, fitsimage)

        self.layertag = 'pixtable-canvas'
        self.pan2mark = False

        prefs = self.fv.get_preferences()
        self.settings = prefs.create_category('plugin_PixTable')
        self.settings.add_defaults(fontsize=12,
                                   font='fixed',
                                   mark_radius=10,
                                   mark_style='cross',
                                   mark_color='lightgreen',
                                   select_color='cyan',
                                   drag_update=True)
        self.settings.load(onError='silent')

        self.dc = self.fv.get_draw_classes()
        canvas = self.dc.DrawingCanvas()
        canvas.set_callback('cursor-changed', self.cursor_cb)
        canvas.enable_draw(True)
        canvas.set_drawtype('point', color='cyan', linestyle='dash')
        canvas.set_callback('draw-event', self.draw_cb)
        canvas.enable_edit(True)
        canvas.set_callback('edit-event', self.edit_cb)
        canvas.add_draw_mode('move', down=self.btndown_cb,
                             move=self.motion_cb, up=self.btnup_cb,
                             key=self.keydown_cb)
        canvas.set_draw_mode('move')
        canvas.register_for_cursor_drawing(self.fitsimage)
        canvas.set_surface(self.fitsimage)
        self.canvas = canvas

        # For pixel table
        self.pixtbl_radius = 2
        self.txt_arr = None
        self.sum_arr = None
        self.sizes = [1, 2, 3, 4]
        self.maxdigits = 9
        self.fmt_cell = '{:> %d.%dg}' % (self.maxdigits - 1, self.maxdigits // 2)
        self.lastx = 0
        self.lasty = 0
        self.font = self.settings.get('font', 'fixed')
        self.fontsize = self.settings.get('fontsize', 12)
        self.fontsizes = [6, 8, 9, 10, 11, 12, 14, 16, 18, 24, 28, 32]
        self.pixview = None
        self._wd = 400
        self._ht = 300
        # hack to set a reasonable starting position for the splitter
        _sz = max(self._wd, self._ht)
        self._split_sizes = [_sz, _sz]
        self.gui_up = False

        # For "marks" feature
        self.mark_radius = self.settings.get('mark_radius', 10)
        self.mark_style = self.settings.get('mark_style', 'cross')
        self.mark_color = self.settings.get('mark_color', 'lightgreen')
        self.select_color = self.settings.get('select_color', 'cyan')
        self.marks = ['None']
        self.mark_index = 0
        self.mark_selected = None
        self.drag_update = self.settings.get('drag_update', True)

    def build_gui(self, container):
        top = Widgets.VBox()
        top.set_border_width(4)

        box, sw, orientation = Widgets.get_oriented_box(container)
        box.set_border_width(4)
        box.set_spacing(2)

        fr = Widgets.Frame("Pixel Values")

        # We just use a ginga widget to implement the pixtable
        pixview = Viewers.CanvasView(logger=self.logger)
        pixview.set_desired_size(self._wd, self._ht)
        bg = colors.lookup_color('#202030')
        pixview.set_bg(*bg)

        bd = pixview.get_bindings()
        bd.enable_zoom(True)
        bd.enable_pan(True)

        self.pixview = pixview
        self.pix_w = Viewers.GingaViewerWidget(pixview)
        fr.set_widget(self.pix_w)
        self.pix_w.resize(self._wd, self._ht)

        paned = Widgets.Splitter(orientation=orientation)
        self.w.splitter = paned
        paned.add_widget(fr)

        self._rebuild_table()

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(4)

        cbox1 = Widgets.ComboBox()
        index = 0
        for i in self.sizes:
            j = 1 + i * 2
            name = "%dx%d" % (j, j)
            cbox1.append_text(name)
            index += 1
        index = self.sizes.index(self.pixtbl_radius)
        cbox1.set_index(index)
        cbox1.add_callback('activated', self.set_cutout_size_cb)
        cbox1.set_tooltip("Select size of pixel table")
        btns.add_widget(cbox1, stretch=0)

        # control for selecting a mark
        cbox2 = Widgets.ComboBox()
        for tag in self.marks:
            cbox2.append_text(tag)
        if self.mark_selected is None:
            cbox2.set_index(0)
        else:
            cbox2.show_text(self.mark_selected)
        cbox2.add_callback('activated', self.mark_select_cb)
        self.w.marks = cbox2
        cbox2.set_tooltip("Select a mark")
        #cbox2.setMinimumContentsLength(8)
        btns.add_widget(cbox2, stretch=0)

        btn1 = Widgets.Button("Delete")
        btn1.add_callback('activated', lambda w: self.clear_mark_cb())
        btn1.set_tooltip("Delete selected mark")
        btn1.set_enabled(len(self.marks) > 1)
        self.w.btn_delete = btn1
        btns.add_widget(btn1, stretch=0)

        btn2 = Widgets.Button("Delete All")
        btn2.add_callback('activated', lambda w: self.clear_all())
        btn2.set_tooltip("Clear all marks")
        btns.add_widget(btn2, stretch=0)
        btn2.set_enabled(len(self.marks) > 1)
        self.w.btn_delete_all = btn2
        btns.add_widget(Widgets.Label(''), stretch=1)

        vbox2 = Widgets.VBox()
        vbox2.add_widget(btns, stretch=0)

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(4)

        btn3 = Widgets.CheckBox("Pan to mark")
        btn3.set_state(self.pan2mark)
        btn3.add_callback('activated', self.pan2mark_cb)
        btn3.set_tooltip("Pan follows selected mark")
        btns.add_widget(btn3)
        btns.add_widget(Widgets.Label(''), stretch=1)

        vbox2.add_widget(btns, stretch=0)

        captions = [
            ('Font size:', 'label', 'Font size', 'combobox',
             'Caption:', 'label', 'Caption', 'entry'),
        ]
        w, b = Widgets.build_info(captions)
        self.w.update(b)
        vbox2.add_widget(w, stretch=0)

        b.font_size.set_tooltip("Set font size for pixel display")
        for size in self.fontsizes:
            b.font_size.append_text(str(size))
        b.font_size.show_text(str(self.fontsize))
        b.font_size.add_callback('activated', self.set_font_size_cb)

        b.caption.set_tooltip("Text to append to the marker")

        vbox2.add_widget(Widgets.Label(''), stretch=1)
        box.add_widget(vbox2, stretch=1)

        ## spacer = Widgets.Label('')
        ## box.add_widget(spacer, stretch=1)

        paned.add_widget(sw)
        paned.set_sizes(self._split_sizes)

        top.add_widget(paned, stretch=1)

        mode = self.canvas.get_draw_mode()
        hbox = Widgets.HBox()
        btn1 = Widgets.RadioButton("Move")
        btn1.set_state(mode == 'move')
        btn1.add_callback('activated',
                          lambda w, val: self.set_mode_cb('move', val))
        btn1.set_tooltip("Choose this to add or move a mark")
        self.w.btn_move = btn1
        hbox.add_widget(btn1)

        btn2 = Widgets.RadioButton("Draw", group=btn1)
        btn2.set_state(mode == 'draw')
        btn2.add_callback('activated',
                          lambda w, val: self.set_mode_cb('draw', val))
        btn2.set_tooltip("Choose this to draw a new or replacement mark")
        self.w.btn_draw = btn2
        hbox.add_widget(btn2)

        btn3 = Widgets.RadioButton("Edit", group=btn1)
        btn3.set_state(mode == 'edit')
        btn3.add_callback('activated',
                          lambda w, val: self.set_mode_cb('edit', val))
        btn3.set_tooltip("Choose this to edit a mark")
        self.w.btn_edit = btn3
        hbox.add_widget(btn3)

        hbox.add_widget(Widgets.Label(''), stretch=1)
        top.add_widget(hbox, stretch=0)

        btns = Widgets.HBox()
        btns.set_border_width(4)
        btns.set_spacing(4)

        btn = Widgets.Button("Close")
        btn.add_callback('activated', lambda w: self.close())
        btns.add_widget(btn)
        btn = Widgets.Button("Help")
        btn.add_callback('activated', lambda w: self.help())
        btns.add_widget(btn, stretch=0)
        btns.add_widget(Widgets.Label(''), stretch=1)

        top.add_widget(btns, stretch=0)
        container.add_widget(top, stretch=1)
        self.gui_up = True

    def select_mark(self, tag, pan=True):
        # deselect the current selected mark, if there is one
        if self.mark_selected is not None:
            try:
                obj = self.canvas.get_object_by_tag(self.mark_selected)
                obj.set_attr_all(color=self.mark_color)
            except Exception:
                # old object may have been deleted
                pass

        self.mark_selected = tag
        if tag is None:
            self.w.marks.show_text('None')
            self.canvas.redraw(whence=3)
            return

        self.w.marks.show_text(tag)
        obj = self.canvas.get_object_by_tag(tag)
        obj.set_attr_all(color=self.select_color)
        self.lastx = obj.objects[0].x
        self.lasty = obj.objects[0].y
        if self.pan2mark and pan:
            self.fitsimage.panset_xy(self.lastx, self.lasty)
        self.canvas.redraw(whence=3)

        self.redo()

    def mark_select_cb(self, w, index):
        tag = self.marks[index]
        if index == 0:
            tag = None
        self.select_mark(tag)

    def pan2mark_cb(self, w, val):
        self.pan2mark = val

    def clear_mark_cb(self):
        tag = self.mark_selected
        if tag is None:
            return
        self.canvas.delete_object_by_tag(tag)
        self.w.marks.delete_alpha(tag)
        self.marks.remove(tag)
        self.w.marks.set_index(0)
        self.mark_selected = None
        self.w.btn_delete.set_enabled(len(self.marks) > 1)
        self.w.btn_delete_all.set_enabled(len(self.marks) > 1)

    def clear_all(self):
        self.canvas.delete_all_objects()
        for name in self.marks:
            self.w.marks.delete_alpha(name)
        self.marks = ['None']
        self.w.marks.append_text('None')
        self.w.marks.set_index(0)
        self.mark_selected = None
        self.mark_index = 0
        self.w.btn_delete.set_enabled(False)
        self.w.btn_delete_all.set_enabled(False)

    def set_font_size_cb(self, w, index):
        self.fontsize = self.fontsizes[index]
        self._rebuild_table()

    def plot(self, data, x1, y1, x2, y2, data_x, data_y, radius,
             maxv=9):

        # Because most FITS data is stored with lower Y indexes to
        # bottom
        data = np.flipud(data)

        width, height = self.fitsimage.get_dims(data)
        if self.txt_arr is None:
            return

        maxval = np.nanmax(data)
        minval = np.nanmin(data)
        avgval = np.mean(data)
        rmsval = np.sqrt(np.mean(np.square(data)))
        fmt_cell = self.fmt_cell

        # can we do this with a np.vectorize() fn call and
        # speed things up?
        for i in range(width):
            for j in range(height):
                val = data[i][j]
                if not np.isscalar(val):
                    val = np.average(val)
                self.txt_arr[i][j].text = fmt_cell.format(val)

        ctr_txt = self.txt_arr[width // 2][height // 2]

        # append statistics line
        fmt_stat = "  Min: %s  Max: %s  Avg: %s  Rms: %s" % (
            fmt_cell, fmt_cell, fmt_cell, fmt_cell)
        self.sum_arr[0].text = fmt_stat.format(minval, maxval, avgval, rmsval)

        # update the pixtable
        self.pixview.panset_xy(ctr_txt.x, ctr_txt.y)
        #self.pixview.redraw(whence=3)

    def close(self):
        self.fv.stop_local_plugin(self.chname, str(self))
        return True

    def start(self):
        # insert layer if it is not already
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.get_object_by_tag(self.layertag)

        except KeyError:
            # Add canvas layer
            p_canvas.add(self.canvas, tag=self.layertag)
        self.resume()

    def stop(self):
        self.gui_up = False
        self._split_sizes = self.w.splitter.get_sizes()
        # remove the canvas from the image
        self.canvas.ui_set_active(False)
        p_canvas = self.fitsimage.get_canvas()
        try:
            p_canvas.delete_object_by_tag(self.layertag)
        except Exception:
            pass
        self.pixview = None

    def pause(self):
        self.canvas.ui_set_active(False)

    def resume(self):
        # turn off any mode user may be in
        self.modes_off()

        self.canvas.ui_set_active(True)
        self.redo()

    def redo(self):
        if self.pixview is None:
            return
        # cut out and set the pixel table data
        image = self.fitsimage.get_image()

        if image is None:
            return

        # We report the value across the pixel, even though the coords
        # change halfway across the pixel
        data_x, data_y = int(self.lastx + 0.5), int(self.lasty + 0.5)

        # cutout image data
        data, x1, y1, x2, y2 = image.cutout_radius(data_x, data_y,
                                                   self.pixtbl_radius)
        self.fv.error_wrap(self.plot, data, x1, y1, x2, y2,
                           self.lastx, self.lasty,
                           self.pixtbl_radius, maxv=9)

    def _rebuild_table(self):
        canvas = self.pixview.get_canvas()
        canvas.delete_all_objects(redraw=False)

        Text = canvas.get_draw_class('text')
        ex_txt = Text(0, 0, text='5', fontsize=self.fontsize, font=self.font)
        font_wd, font_ht = self.fitsimage.renderer.get_dimensions(ex_txt)
        max_wd = self.maxdigits + 2
        crdmap = self.pixview.get_coordmap('window')

        rows = []
        objs = []
        max_x = 0
        for row in range(self.pixtbl_radius * 2 + 1):
            cols = []
            for col in range(self.pixtbl_radius * 2 + 1):
                col_wd = font_wd * max_wd
                x = col_wd * col + 4
                max_x = max(max_x, x + col_wd)
                y = font_ht * (row + 1) + 4

                color = 'lightgreen'
                if (row == col) and (row == self.pixtbl_radius):
                    color = 'pink'

                dx, dy = crdmap.to_data((x, y))
                text_obj = Text(dx, dy, text='', font=self.font,
                                color=color, fontsize=self.fontsize,
                                coord='data')
                objs.append(text_obj)
                cols.append(text_obj)

            rows.append(cols)

        self.txt_arr = np.array(rows)

        # add summary row(s)
        x = (font_wd + 2) + 4
        y += font_ht + 20
        dx, dy = crdmap.to_data((x, y))
        s1 = Text(dx, dy, text='', font=self.font,
                  color=color, fontsize=self.fontsize,
                  coord='data')
        objs.append(s1)
        self.sum_arr = np.array([s1])

        # add all of the text objects to the canvas as one large
        # compound object
        CompoundObject = canvas.get_draw_class('compoundobject')
        canvas.add(CompoundObject(*objs), redraw=False)

        # set limits for scrolling
        self.pixview.set_limits(((0, 0), (max_x, y)), coord='window')

    def set_cutout_size_cb(self, w, val):
        index = w.get_index()
        self.pixtbl_radius = self.sizes[index]
        self._rebuild_table()

    def cursor_cb(self, canvas, junk, data_x, data_y):
        if self.mark_selected is not None:
            return False
        if self.pixview is None:
            return

        self.lastx, self.lasty = data_x, data_y

        self.redo()
        return False

    def add_mark(self, data_x, data_y, radius=None, color=None, style=None,
                 text=None):
        if not radius:
            radius = self.mark_radius
        if not color:
            color = self.mark_color
        if not style:
            style = self.mark_style

        self.logger.debug("Setting mark at %d,%d" % (data_x, data_y))
        self.mark_index += 1
        tag = 'mark%d' % (self.mark_index)
        caption = "%d" % (self.mark_index)
        if text is not None:
            caption = caption + ': ' + text
        if radius is None:
            radius = self.mark_radius
        pt_obj = self.dc.Point(data_x, data_y, radius,
                               style=style, color=color,
                               linestyle='solid')
        txt_obj = self.dc.Text(10, 0, caption,
                               color=color, ref_obj=pt_obj, coord='offset')
        txt_obj.editable = False
        tag = self.canvas.add(self.dc.CompoundObject(pt_obj, txt_obj),
                              tag=tag)
        self.marks.append(tag)
        self.w.marks.append_text(tag)
        self.w.btn_delete.set_enabled(True)
        self.w.btn_delete_all.set_enabled(True)
        self.select_mark(tag, pan=False)

    def _mark_update(self, data_x, data_y):
        if self.mark_selected is None:
            return False

        m_obj = self.canvas.get_object_by_tag(self.mark_selected)
        p_obj = m_obj.objects[0]
        p_obj.move_to(data_x, data_y)
        self.lastx, self.lasty = data_x, data_y
        self.canvas.update_canvas()
        return True

    def btndown_cb(self, canvas, event, data_x, data_y, viewer):
        if self._mark_update(data_x, data_y):
            if self.drag_update:
                self.redo()
            return True

        # no selected mark, make a new one
        caption = self.w.caption.get_text().strip()
        if len(caption) == 0:
            caption = None
        self.add_mark(data_x, data_y, text=caption)
        return True

    def motion_cb(self, canvas, event, data_x, data_y, viewer):
        if not self._mark_update(data_x, data_y):
            return False

        if self.drag_update:
            self.redo()
        return True

    def btnup_cb(self, canvas, event, data_x, data_y, viewer):
        if not self._mark_update(data_x, data_y):
            return False

        self.redo()
        return True

    def prev_mark(self):
        if len(self.marks) <= 1 or self.mark_selected is None:
            # no previous
            return

        idx = self.marks.index(self.mark_selected)
        idx = idx - 1
        if idx < 0:
            return
        tag = self.marks[idx]
        if tag == 'None':
            tag = None
        self.select_mark(tag)

    def next_mark(self):
        if len(self.marks) <= 1:
            # no next
            return

        if self.mark_selected is None:
            idx = 0
        else:
            idx = self.marks.index(self.mark_selected)
        idx = idx + 1
        if idx >= len(self.marks):
            return
        tag = self.marks[idx]
        if tag == 'None':
            tag = None
        self.select_mark(tag)

    def keydown_cb(self, canvas, event, data_x, data_y, viewer):
        if event.key == 'n':
            caption = self.w.caption.get_text().strip()
            if len(caption) == 0:
                caption = None
            self.add_mark(data_x, data_y, text=caption)
            return True
        elif event.key == 'm':
            if self._mark_update(data_x, data_y):
                self.redo()
            return True
        elif event.key == 'd':
            self.clear_mark_cb()
            return True
        elif event.key == 'j':
            self.prev_mark()
            return True
        elif event.key == 'k':
            self.next_mark()
            return True
        return False

    def draw_cb(self, canvas, tag):
        obj = canvas.get_object_by_tag(tag)
        canvas.delete_object_by_tag(tag)

        caption = self.w.caption.get_text().strip()
        if len(caption) == 0:
            caption = None
        self.add_mark(obj.x, obj.y, text=caption, radius=obj.radius)

    def edit_cb(self, canvas, obj):
        if self.mark_selected is not None:
            m_obj = self.canvas.get_object_by_tag(self.mark_selected)
            if m_obj is not None and m_obj.objects[0] is obj:
                # edited mark was the selected mark
                self.lastx, self.lasty = obj.x, obj.y
                self.redo()
        return True

    def edit_select_mark(self):
        if self.mark_selected is not None:
            obj = self.canvas.get_object_by_tag(self.mark_selected)
            # drill down to reference shape
            if hasattr(obj, 'objects'):
                obj = obj.objects[0]
            self.canvas.edit_select(obj)
        else:
            self.canvas.clear_selected()
        self.canvas.update_canvas()

    def set_mode_cb(self, mode, tf):
        """Called when one of the Move/Draw/Edit radio buttons is selected."""
        if tf:
            self.canvas.set_draw_mode(mode)
            if mode == 'edit':
                self.edit_select_mark()
        return True

    def set_mode(self, mode):
        self.canvas.set_draw_mode(mode)
        self.w.btn_move.set_state(mode == 'move')
        self.w.btn_draw.set_state(mode == 'draw')
        self.w.btn_edit.set_state(mode == 'edit')

    def __str__(self):
        return 'pixtable'


# Append module docstring with config doc for auto insert by Sphinx.
from ginga.util.toolbox import generate_cfg_example  # noqa
if __doc__ is not None:
    __doc__ += generate_cfg_example('plugin_PixTable', package='ginga')

# END



try:
    # As long as we're using setuptools/distribute, we need to do this the
    # setuptools way or else pkg_resources will throw up unncessary and
    # annoying warnings (even though the namespace mechanism will still
    # otherwise work without it).
    # Get rid of this as soon as setuptools/distribute is dead.
    __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
    pass
__path__ = __import__('pkgutil').extend_path(__path__, __name__)



#
# iqcalc.py -- image quality calculations on FITS data
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#

import math
import logging
import threading

import numpy as np

try:
    import scipy.optimize as optimize
    import scipy.ndimage as ndimage
    import scipy.ndimage.filters as filters
    have_scipy = True
except ImportError:
    have_scipy = False

from ginga.misc import Bunch


def get_mean(data_np):
    """Calculate mean for valid values.

    Parameters
    ----------
    data_np : ndarray
        Input array.

    Returns
    -------
    result : float
        Mean of array values that are finite.
        If array contains no finite values, returns NaN.

    """
    i = np.isfinite(data_np)
    if not np.any(i):
        return np.nan
    return np.mean(data_np[i])


def get_median(data_np):
    """Like :func:`get_mean` but for median."""
    i = np.isfinite(data_np)
    if not np.any(i):
        return np.nan
    return np.median(data_np[i])


class IQCalcError(Exception):
    """Base exception for raising errors in this module."""
    pass


class IQCalc(object):

    def __init__(self, logger=None):
        if not logger:
            logger = logging.getLogger('IQCalc')
        self.logger = logger

        # for mutex around scipy.optimize, which seems to be non-threadsafe
        self.lock = threading.RLock()

        # for adjustments to background level
        self.skylevel_magnification = 1.05
        self.skylevel_offset = 40.0

    # FWHM CALCULATION

    def gaussian(self, x, p):
        """Gaussian fitting function in 1D.  Makes a sine function with
        amplitude determined by maxv.  See calc_fwhm().

        p[0]==mean, p[1]==sdev, p[2]=maxv
        """
        y = (1.0 / (p[1] * np.sqrt(2 * np.pi)) *
             np.exp(-(x - p[0]) ** 2 / (2 * p[1] ** 2))) * p[2]
        return y

    def calc_fwhm_gaussian(self, arr1d, medv=None, gauss_fn=None):
        """FWHM calculation on a 1D array by using least square fitting of
        a gaussian function on the data.  arr1d is a 1D array cut in either
        X or Y direction on the object.
        """
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")
        if gauss_fn is None:
            gauss_fn = self.gaussian

        N = len(arr1d)
        X = np.array(list(range(N)))
        Y = arr1d
        # Fitting works more reliably if we do the following
        # a. subtract sky background
        if medv is None:
            medv = get_median(Y)
        Y = Y - medv
        maxv = Y.max()
        # b. clamp to 0..max (of the sky subtracted field)
        Y = Y.clip(0, maxv)

        # Fit a gaussian
        p0 = [0, N - 1, maxv]              # Inital guess
        # Distance to the target function
        errfunc = lambda p, x, y: gauss_fn(x, p) - y  # noqa
        # Least square fit to the gaussian
        with self.lock:
            # NOTE: without this mutex, optimize.leastsq causes a fatal error
            # sometimes--it appears not to be thread safe.
            # The error is:
            # "SystemError: null argument to internal routine"
            # "Fatal Python error: GC object already tracked"
            p1, success = optimize.leastsq(errfunc, p0[:], args=(X, Y))

        if not success:
            raise IQCalcError("FWHM gaussian fitting failed")

        mu, sdev, maxv = p1
        self.logger.debug("mu=%f sdev=%f maxv=%f" % (mu, sdev, maxv))

        # Now that we have the sdev from fitting, we can calculate FWHM
        fwhm = 2.0 * np.sqrt(2.0 * np.log(2.0)) * sdev
        # some routines choke on numpy values and need "pure" Python floats
        # e.g. when marshalling through a remote procedure interface
        fwhm = float(fwhm)
        mu = float(mu)
        sdev = float(sdev)
        maxv = float(maxv)

        res = Bunch.Bunch(fwhm=fwhm, mu=mu, sdev=sdev, maxv=maxv,
                          fit_fn=gauss_fn, fit_args=[mu, sdev, maxv])
        return res

    def moffat(self, x, p):
        """Moffat fitting function in 1D.
        p[0]==mean, p[1]==corewidth(gamma), p[2]=power(alpha), p[3]=maxv
        """
        y = (1.0 + (x - p[0]) ** 2 / p[1] ** 2) ** (-1.0 * p[2]) * p[3]
        return y

    def calc_fwhm_moffat(self, arr1d, medv=None, moffat_fn=None):
        """FWHM calculation on a 1D array by using least square fitting of
        a Moffat function on the data.  arr1d is a 1D array cut in either
        X or Y direction on the object.
        """
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")
        if moffat_fn is None:
            moffat_fn = self.moffat

        N = len(arr1d)
        X = np.array(list(range(N)))
        Y = arr1d
        # Fitting works more reliably if we do the following
        # a. subtract sky background
        if medv is None:
            medv = get_median(Y)
        Y = Y - medv
        maxv = Y.max()
        # b. clamp to 0..max (of the sky subtracted field)
        Y = Y.clip(0, maxv)

        # Fit a moffat
        p0 = [0, N - 1, 2, maxv]              # Inital guess
        # Distance to the target function
        errfunc = lambda p, x, y: moffat_fn(x, p) - y  # noqa
        # Least square fit to the gaussian
        with self.lock:
            # NOTE: without this mutex, optimize.leastsq causes a fatal error
            # sometimes--it appears not to be thread safe.
            # The error is:
            # "SystemError: null argument to internal routine"
            # "Fatal Python error: GC object already tracked"
            p1, success = optimize.leastsq(errfunc, p0[:], args=(X, Y))

        if not success:
            raise IQCalcError("FWHM moffat fitting failed")

        mu, width, power, maxv = p1
        width = np.abs(width)
        self.logger.debug("mu=%f width=%f power=%f maxv=%f" % (
            mu, width, power, maxv))

        fwhm = 2.0 * width * np.sqrt(2.0 ** (1.0 / power) - 1.0)

        # some routines choke on numpy values and need "pure" Python floats
        # e.g. when marshalling through a remote procedure interface
        fwhm = float(fwhm)
        mu = float(mu)
        width = float(width)
        power = float(power)
        maxv = float(maxv)

        res = Bunch.Bunch(fwhm=fwhm, mu=mu, width=width, power=power,
                          maxv=maxv, fit_fn=moffat_fn,
                          fit_args=[mu, width, power, maxv])
        return res

    def calc_fwhm(self, arr1d, medv=None, method_name='gaussian'):

        # Calculate FWHM in each direction
        fwhm_fn = self.calc_fwhm_gaussian
        if method_name == 'moffat':
            fwhm_fn = self.calc_fwhm_moffat

        return fwhm_fn(arr1d, medv=medv)

    def get_fwhm(self, x, y, radius, data, medv=None, method_name='gaussian'):
        """Get the FWHM value of the object at the coordinates (x, y) using
        radius.
        """
        if medv is None:
            medv = get_median(data)

        # Get two cuts of the data, one in X and one in Y
        x0, y0, xarr, yarr = self.cut_cross(x, y, radius, data)

        # Calculate FWHM in each direction
        x_res = self.calc_fwhm(xarr, medv=medv, method_name=method_name)
        fwhm_x, cx = x_res.fwhm, x_res.mu

        y_res = self.calc_fwhm(yarr, medv=medv, method_name=method_name)
        fwhm_y, cy = y_res.fwhm, y_res.mu

        ctr_x = x0 + cx
        ctr_y = y0 + cy
        self.logger.debug("fwhm_x,fwhm_y=%f,%f center=%f,%f" % (
            fwhm_x, fwhm_y, ctr_x, ctr_y))
        return (fwhm_x, fwhm_y, ctr_x, ctr_y, x_res, y_res)

    def starsize(self, fwhm_x, deg_pix_x, fwhm_y, deg_pix_y):
        cdelta1 = math.fabs(deg_pix_x)
        cdelta2 = math.fabs(deg_pix_y)
        fwhm = (fwhm_x * cdelta1 + fwhm_y * cdelta2) / 2.0
        fwhm = fwhm * 3600.0
        return fwhm

    def centroid(self, data, xc, yc, radius):
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")
        xc, yc = int(xc), int(yc)
        x0, y0, arr = self.cut_region(xc, yc, int(radius), data)
        # See https://stackoverflow.com/questions/25369982/center-of-mass-for-roi-in-python
        cp_arr = np.asarray(arr)
        cy, cx = ndimage.center_of_mass(cp_arr)
        return (x0 + cx, y0 + cy)

    # FINDING BRIGHT PEAKS

    def get_threshold(self, data, sigma=5.0):
        # remove masked elements
        fdata = data[np.logical_not(np.ma.getmaskarray(data))]
        # remove Inf or NaN
        fdata = fdata[np.isfinite(fdata)]

        # find the median
        median = get_median(fdata)

        # NOTE: for this method a good default sigma is 5.0
        dist = np.fabs(fdata - median).mean()
        threshold = median + sigma * dist

        # NOTE: for this method a good default sigma is 2.0
        ## std = np.std(fdata - median)
        ## threshold = median + sigma * std

        self.logger.debug("calc threshold=%f" % (threshold))
        return threshold

    def find_bright_peaks(self, data, threshold=None, sigma=5, radius=5):
        """
        Find bright peak candidates in (data).  (threshold) specifies a
        threshold value below which an object is not considered a candidate.
        If threshold is blank, a default is calculated using (sigma).
        (radius) defines a pixel radius for determining local maxima--if the
        desired objects are larger in size, specify a larger radius.

        The routine returns a list of candidate object coordinate tuples
        (x, y) in data.
        """
        if not have_scipy:
            raise IQCalcError("Please install the 'scipy' module "
                              "to use this function")
        if threshold is None:
            # set threshold to default if none provided
            threshold = self.get_threshold(data, sigma=sigma)
            self.logger.debug("threshold defaults to %f (sigma=%f)" % (
                threshold, sigma))

        #self.logger.debug("filtering")
        data_max = filters.maximum_filter(data, radius)
        maxima = (data == data_max)
        diff = data_max > threshold
        maxima[diff == 0] = 0

        #self.logger.debug("finding")
        labeled, num_objects = ndimage.label(maxima)
        slices = ndimage.find_objects(labeled)
        peaks = []
        for dy, dx in slices:
            xc = (dx.start + dx.stop - 1) / 2.0
            yc = (dy.start + dy.stop - 1) / 2.0

            # This is only an approximate center; use FWHM or centroid
            # calculation to refine further
            peaks.append((xc, yc))

        self.logger.debug("peaks=%s" % (str(peaks)))
        return peaks

    def cut_region(self, x, y, radius, data):
        """Return a cut region (radius) pixels away from (x, y) in (data).
        """
        n = radius
        ht, wd = data.shape
        x0, x1 = max(0, x - n), min(wd - 1, x + n)
        y0, y1 = max(0, y - n), min(ht - 1, y + n)
        arr = data[y0:y1 + 1, x0:x1 + 1]
        return (x0, y0, arr)

    def cut_cross(self, x, y, radius, data):
        """Cut two data subarrays that have a center at (x, y) and with
        radius (radius) from (data).  Returns the starting pixel (x0, y0)
        of each cut and the respective arrays (xarr, yarr).
        """
        n = int(round(radius))
        ht, wd = data.shape
        x, y = int(round(x)), int(round(y))
        x0, x1 = int(max(0, x - n)), int(min(wd - 1, x + n))
        y0, y1 = int(max(0, y - n)), int(min(ht - 1, y + n))
        xarr = data[y, x0:x1 + 1]
        yarr = data[y0:y1 + 1, x]
        return (x0, y0, xarr, yarr)

    def brightness(self, x, y, radius, medv, data):
        """Return the brightness value found in a region (radius) pixels away
        from (x, y) in (data).
        """
        x0, y0, arr = self.cut_region(x, y, radius, data)
        arr2 = np.sort(arr.flat)
        idx = int(len(arr2) * 0.8)
        res = arr2[idx] - medv
        return float(res)

    def fwhm_data(self, x, y, data, radius=15, method_name='gaussian'):
        return self.get_fwhm(x, y, radius, data, method_name=method_name)

    # EVALUATION ON A FIELD

    def evaluate_peaks(self, peaks, data, bright_radius=2, fwhm_radius=15,
                       fwhm_method='gaussian', cb_fn=None, ev_intr=None):

        height, width = data.shape
        hh = float(height) / 2.0
        ht = float(height)
        h4 = float(height) * 4.0
        wh = float(width) / 2.0
        wd = float(width)
        w4 = float(width) * 4.0

        # Find the median (sky/background) level
        median = float(get_median(data))
        #skylevel = median
        # Old SOSS qualsize() applied this calculation to skylevel
        skylevel = median * self.skylevel_magnification + self.skylevel_offset

        # Form a list of objects and their characteristics
        objlist = []
        for x, y in peaks:
            if ev_intr and ev_intr.is_set():
                raise IQCalcError("Evaluation interrupted!")

            # Find the fwhm in x and y
            try:
                res = self.fwhm_data(x, y, data, radius=fwhm_radius,
                                     method_name=fwhm_method)
                fwhm_x, fwhm_y, ctr_x, ctr_y, x_res, y_res = res

                bx = x_res.fit_fn(round(ctr_x),
                                  (ctr_x,) + tuple(x_res.fit_args[1:]))
                by = y_res.fit_fn(round(ctr_y),
                                  (ctr_y,) + tuple(y_res.fit_args[1:]))
                bright = float((bx + by) / 2.0)

            except Exception as e:
                # Error doing FWHM, skip this object
                self.logger.debug("Error doing FWHM on object at %.2f,%.2f: %s" % (
                    x, y, str(e)))
                continue

            oid_x, oid_y = None, None
            try:
                oid_x, oid_y = self.centroid(data, x, y, fwhm_radius)

            except Exception as e:
                # Error doing centroid
                self.logger.debug("Error doing centroid on object at %.2f,%.2f: %s" % (
                    x, y, str(e)))

            self.logger.debug("orig=%f,%f  ctr=%f,%f  fwhm=%f,%f bright=%f" % (
                x, y, ctr_x, ctr_y, fwhm_x, fwhm_y, bright))
            # overall measure of fwhm as a single value
            fwhm = (math.sqrt(fwhm_x * fwhm_x + fwhm_y * fwhm_y) *
                    (1.0 / math.sqrt(2.0)))

            # calculate a measure of ellipticity
            elipse = math.fabs(min(fwhm_x, fwhm_y) / max(fwhm_x, fwhm_y))

            # calculate a measure of distance from center of image
            dx = wh - ctr_x
            dy = hh - ctr_y
            dx2 = dx * dx / wd / w4
            dy2 = dy * dy / ht / h4
            if dx2 > dy2:
                pos = 1.0 - dx2
            else:
                pos = 1.0 - dy2

            obj = Bunch.Bunch(objx=ctr_x, objy=ctr_y, pos=pos,
                              oid_x=oid_x, oid_y=oid_y,
                              fwhm_x=fwhm_x, fwhm_y=fwhm_y,
                              fwhm=fwhm, fwhm_radius=fwhm_radius,
                              brightness=bright, elipse=elipse,
                              x=int(x), y=int(y),
                              skylevel=skylevel, background=median)
            objlist.append(obj)

            if cb_fn is not None:
                cb_fn(obj)

        return objlist

    def _sortkey(self, obj):
        val = obj.brightness * obj.pos / math.sqrt(obj.fwhm)
        return val

    def objlist_select(self, objlist, width, height,
                       minfwhm=2.0, maxfwhm=150.0, minelipse=0.5,
                       edgew=0.01):

        results = []
        count = 0
        for obj in objlist:
            count += 1
            self.logger.debug("%d obj x,y=%.2f,%.2f fwhm=%.2f bright=%.2f" % (
                count, obj.objx, obj.objy, obj.fwhm, obj.brightness))
            # If peak has a minfwhm < fwhm < maxfwhm and the object
            # is inside the frame by edgew pct
            if ((minfwhm < obj.fwhm) and (obj.fwhm < maxfwhm) and
                    (minelipse < obj.elipse) and (width * edgew < obj.x) and
                    (height * edgew < obj.y) and
                    (width * (1.0 - edgew) > obj.x) and
                    (height * (1.0 - edgew) > obj.y)):
                results.append(obj)

        #results.sort(cmp=self._compare)
        results.sort(key=self._sortkey, reverse=True)
        return results

    def pick_field(self, data, peak_radius=5, bright_radius=2, fwhm_radius=15,
                   threshold=None,
                   minfwhm=2.0, maxfwhm=50.0, minelipse=0.5,
                   edgew=0.01):

        height, width = data.shape

        # Find the bright peaks in the image
        peaks = self.find_bright_peaks(data, radius=peak_radius,
                                       threshold=threshold)
        #print "peaks=", peaks
        self.logger.debug("peaks=%s" % str(peaks))
        if len(peaks) == 0:
            raise IQCalcError("Cannot find bright peaks")

        # Evaluate those peaks
        objlist = self.evaluate_peaks(peaks, data,
                                      bright_radius=bright_radius,
                                      fwhm_radius=fwhm_radius)
        if len(objlist) == 0:
            raise IQCalcError("Error evaluating bright peaks")

        results = self.objlist_select(objlist, width, height,
                                      minfwhm=minfwhm, maxfwhm=maxfwhm,
                                      minelipse=minelipse, edgew=edgew)
        if len(results) == 0:
            raise IQCalcError("No object matches selection criteria")

        return results[0]

    def qualsize(self, image, x1=None, y1=None, x2=None, y2=None,
                 radius=5, bright_radius=2, fwhm_radius=15, threshold=None,
                 minfwhm=2.0, maxfwhm=50.0, minelipse=0.5,
                 edgew=0.01):

        x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
        data = image.cutout_data(x1, y1, x2, y2, astype=np.float)

        qs = self.pick_field(data, peak_radius=radius,
                             bright_radius=bright_radius,
                             fwhm_radius=fwhm_radius,
                             threshold=threshold,
                             minfwhm=minfwhm, maxfwhm=maxfwhm,
                             minelipse=minelipse, edgew=edgew)

        # Add back in offsets into image to get correct values with respect
        # to the entire image
        qs.x += x1
        qs.y += y1
        qs.objx += x1
        qs.objy += y1
        self.logger.debug("obj=%f,%f fwhm=%f sky=%f bright=%f" % (
            qs.objx, qs.objy, qs.fwhm, qs.skylevel, qs.brightness))

        return qs

#END



#
# CairoHelp.py -- help classes for the cairo-based operations
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.

import itertools

import numpy as np

have_cairo = False
try:
    import cairo
    have_cairo = True
except ImportError:
    pass


def text_size(text, font):
    """Calculate the size of text at a given font.

    Parameters
    ----------
    text : str
        A text string

    font : `~ginga.opengl.GlHelp.Font` or compatible object
        A Ginga font descriptor

    Returns
    -------
    (wd, ht) : tuple of int
        Size the text would occupy in pixels.
    """
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
    ctx = cairo.Context(surface)
    ctx.select_font_face(font.fontname, cairo.FONT_SLANT_NORMAL)
    ctx.set_font_size(font.fontsize)
    tup = ctx.text_extents(text)
    wd, ht = int(round(tup[2])), int(round(tup[3]))
    return wd, ht


def text_to_paths(text, font, cx=0, cy=0, rot_deg=0.0, flip_y=False):
    """Convert a text string into paths.

    Parameters
    ----------
    text : str
        A text string

    font :
        A Ginga font descriptor

    cx : int (optional, default 0)
        X offset to add to each path point generated

    cy : int (optional, default 0)
        Y offset to add to each path point generated

    rot_deg : float (optional, default 0.0)
        Amount to rotate the text counterclockwise

    flip_y : bool (optional, default False)
        Whether to flip the coordinates in Y before

    Returns
    -------
    paths : list of ndarray
        A list of paths represented by numpy arrays of (x, y) points;
        the paths can be drawn to render the text
    """
    wd, ht = text_size(text, font)

    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, wd, ht)
    ctx = cairo.Context(surface)
    ctx.move_to(0, 0)
    ctx.rotate(-np.radians(rot_deg))

    ctx.select_font_face(font.fontname, cairo.FONT_SLANT_NORMAL)
    ctx.set_font_size(font.fontsize)

    ctx.text_path(text)

    p = ctx.copy_path_flat()
    c_paths = [pt if i != cairo.PATH_CLOSE_PATH else None for (i, pt) in p]

    paths = []
    for k, g in itertools.groupby(c_paths, key=lambda t: t is not None):
        lg = list(g)
        if lg[0] is not None and len(lg[0]) > 1:
            pts = np.array(lg)
            off = 0
            if flip_y:
                max_y = ht
                pts.T[1] = max_y - pts.T[1]
                off = ht
            pts = np.add(pts, (cx, cy - off))
            paths.append(pts)

    return paths



"""Test ap_regions.py"""

import numpy as np
import pytest

regions = pytest.importorskip('regions')

from astropy import units as u

from ginga.util.ap_region import (astropy_region_to_ginga_canvas_object as r2g,
                                  ginga_canvas_object_to_astropy_region as g2r)
from ginga.canvas.CanvasObject import get_canvas_types


dc = get_canvas_types()


class Test_R2G(object):
    """Test conversions from AstroPy regions to Ginga canvas types."""

    def test_point1(self):
        r = regions.PointPixelRegion(center=regions.PixCoord(x=42, y=43),
                                     visual=regions.RegionVisual(symbol='x'))
        o = r2g(r)
        assert isinstance(o, dc.Point) and o.style == 'cross'
        assert np.all(np.isclose((o.x, o.y), (r.center.x, r.center.y)))

    def test_point2(self):
        r = regions.PointPixelRegion(center=regions.PixCoord(x=42, y=43),
                                     visual=regions.RegionVisual(symbol='+'))
        o = r2g(r)
        assert isinstance(o, dc.Point) and o.style == 'plus'

    def test_point3(self):
        r = regions.PointPixelRegion(center=regions.PixCoord(x=42, y=43),
                                     visual=regions.RegionVisual(symbol='*'))
        o = r2g(r)
        assert isinstance(o, dc.Point) and o.style == 'square'

    def test_text(self):
        r = regions.TextPixelRegion(center=regions.PixCoord(x=42, y=43),
                                    text='Foo',
                                    visual=regions.RegionVisual(textangle='45'))
        o = r2g(r)
        assert isinstance(o, dc.Text) and o.text == 'Foo'
        assert np.all(np.isclose((o.x, o.y, o.rot_deg),
                                 (r.center.x, r.center.y, float(r.visual['textangle']))))

    def test_line(self):
        r = regions.LinePixelRegion(start=regions.PixCoord(x=42, y=43),
                                    end=regions.PixCoord(x=42, y=43))
        o = r2g(r)
        assert isinstance(o, dc.Line)
        assert np.all(np.isclose((o.x1, o.y1, o.x2, o.y2),
                                 (r.start.x, r.start.y, r.end.x, r.end.y)))

    def test_rectangle(self):
        r = regions.RectanglePixelRegion(center=regions.PixCoord(x=42, y=43),
                                         width=3, height=4,
                                         angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Box)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.rot_deg),
                                 (r.center.x, r.center.y, r.width / 2.0, r.height / 2.0,
                                  r.angle.to(u.deg).value)))

    def test_circle(self):
        r = regions.CirclePixelRegion(center=regions.PixCoord(x=42, y=43),
                                      radius=4.2)
        o = r2g(r)
        assert isinstance(o, dc.Circle)
        assert np.all(np.isclose((o.x, o.y, o.radius), (r.center.x, r.center.y, r.radius)))

    def test_ellipse(self):
        r = regions.EllipsePixelRegion(center=regions.PixCoord(x=42, y=43),
                                       height=4.2, width=4.2, angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Ellipse)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.rot_deg),
                                 (r.center.x, r.center.y, r.width / 2., r.height / 2.,
                                  r.angle.to(u.deg).value)))

    def test_polygon(self):
        r = regions.PolygonPixelRegion(vertices=regions.PixCoord(x=[1, 2, 2], y=[1, 1, 2]))
        o = r2g(r)
        assert isinstance(o, dc.Polygon)
        assert np.all(np.isclose(o.points, np.array(r.vertices.xy).T))

    def test_circle_annulus(self):
        r = regions.CircleAnnulusPixelRegion(center=regions.PixCoord(x=42, y=43),
                                             inner_radius=4.2, outer_radius=5.2)
        o = r2g(r)
        assert isinstance(o, dc.Annulus)
        assert np.all(np.isclose((o.x, o.y, o.radius, o.radius + o.width),
                                 (r.center.x, r.center.y, r.inner_radius, r.outer_radius)))

    def test_ellipse_annulus(self):
        r = regions.EllipseAnnulusPixelRegion(center=regions.PixCoord(x=42, y=43),
                                              inner_width=4.2, outer_width=5.2,
                                              inner_height=7.2, outer_height=8.2,
                                              angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Annulus2R) and o.atype == 'ellipse'
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.x, r.center.y, r.inner_width / 2., r.inner_height / 2.,
                                  r.outer_width / 2., r.outer_height / 2.,
                                  r.angle.to(u.deg).value)))

    def test_rectangle_annulus(self):
        r = regions.RectangleAnnulusPixelRegion(center=regions.PixCoord(x=42, y=43),
                                                inner_width=4.2, outer_width=5.2,
                                                inner_height=7.2, outer_height=8.2,
                                                angle=5 * u.deg)
        o = r2g(r)
        assert isinstance(o, dc.Annulus2R) and o.atype == 'box'
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.x, r.center.y, r.inner_width / 2., r.inner_height / 2.,
                                  r.outer_width / 2., r.outer_height / 2.,
                                  r.angle.to(u.deg).value)))


class Test_G2R(object):
    """Test conversions from Ginga canvas types to AstroPy regions."""

    def test_point1(self):
        o = dc.Point(42, 43, radius=5, style='cross')
        r = g2r(o)
        assert isinstance(r, regions.PointPixelRegion)
        assert np.all(np.isclose((o.x, o.y), (r.center.x, r.center.y)))
        assert r.visual['symbol'] == 'x'

    def test_point2(self):
        o = dc.Point(42, 43, radius=5, style='plus')
        r = g2r(o)
        assert isinstance(r, regions.PointPixelRegion)
        assert r.visual['symbol'] == '+'

    def test_point3(self):
        o = dc.Point(42, 43, radius=5, style='square')
        r = g2r(o)
        assert isinstance(r, regions.PointPixelRegion)
        assert r.visual['symbol'] == '*'

    def test_text(self):
        o = dc.Text(42, 43, text='Foo', rot_deg=45.0)
        r = g2r(o)
        assert isinstance(r, regions.TextPixelRegion) and r.text == 'Foo'
        assert np.all(np.isclose((o.x, o.y, o.rot_deg),
                                 (r.center.x, r.center.y, float(r.visual['textangle']))))

    def test_line(self):
        o = dc.Line(42, 43, 52, 53)
        r = g2r(o)
        assert isinstance(r, regions.LinePixelRegion)
        assert np.all(np.isclose((o.x1, o.y1, o.x2, o.y2),
                                 (r.start.x, r.start.y, r.end.x, r.end.y)))

    def test_box(self):
        o = dc.Box(42, 43, 10, 10)
        r = g2r(o)
        assert isinstance(r, regions.RectanglePixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius),
                                 (r.center.x, r.center.y, r.width / 2.0, r.height / 2.0)))

    def test_circle(self):
        o = dc.Circle(42, 43, 4.2)
        r = g2r(o)
        assert isinstance(r, regions.CirclePixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.radius), (r.center.x, r.center.y, r.radius)))

    def test_ellipse(self):
        o = dc.Ellipse(42, 43, 4.2, 4.2, rot_deg=5.0)
        r = g2r(o)
        assert isinstance(r, regions.EllipsePixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.rot_deg),
                                 (r.center.x, r.center.y, r.width / 2., r.height / 2.,
                                  r.angle.to(u.deg).value)))

    def test_polygon(self):
        o = dc.Polygon([(1, 1), (2, 1), (2, 2)])
        r = g2r(o)
        assert isinstance(r, regions.PolygonPixelRegion)
        assert np.all(np.isclose(o.points, np.array(r.vertices.xy).T))

    def test_circle_annulus(self):
        o = dc.Annulus(42, 43, 4.2, width=1, atype='circle')
        r = g2r(o)
        assert isinstance(r, regions.CircleAnnulusPixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.radius, o.radius + o.width),
                                 (r.center.x, r.center.y, r.inner_radius, r.outer_radius)))

    def test_ellipse_annulus(self):
        o = dc.Annulus2R(42, 43, 4.2, 7.2, xwidth=1, ywidth=1, rot_deg=5.0, atype='ellipse')
        r = g2r(o)
        assert isinstance(r, regions.EllipseAnnulusPixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.x, r.center.y, r.inner_width / 2., r.inner_height / 2.,
                                  r.outer_width / 2., r.outer_height / 2.,
                                  r.angle.to(u.deg).value)))

    def test_rectangle_annulus(self):
        o = dc.Annulus2R(42, 43, 4.2, 7.2, xwidth=1, ywidth=1, rot_deg=5.0, atype='box')
        r = g2r(o)
        assert isinstance(r, regions.RectangleAnnulusPixelRegion)
        assert np.all(np.isclose((o.x, o.y, o.xradius, o.yradius, o.xradius + o.xwidth,
                                  o.yradius + o.ywidth, o.rot_deg),
                                 (r.center.x, r.center.y, r.inner_width / 2., r.inner_height / 2.,
                                  r.outer_width / 2., r.outer_height / 2.,
                                  r.angle.to(u.deg).value)))



"""Unit Tests for the cmap.py functions"""

import numpy as np
import pytest

import ginga.cmap
from ginga.cmap import ColorMap


class TestCmap(object):

    def setup_class(self):
        pass

    def test_ColorMap_init(self):
        test_clst = tuple([(x, x, x)
                           for x in np.linspace(0, 1, ginga.cmap.min_cmap_len)])
        test_color_map = ColorMap('test-name', test_clst)

        expected = 'test-name'
        actual = test_color_map.name
        assert expected == actual

        expected = ginga.cmap.min_cmap_len
        actual = len(test_color_map.clst)
        assert expected == actual

        expected = (0.0, 0.0, 0.0)
        actual = test_color_map.clst[0]
        assert np.allclose(expected, actual)

        expected = (1.0, 1.0, 1.0)
        actual = test_color_map.clst[-1]
        assert np.allclose(expected, actual)

    def test_ColorMap_init_exception(self):
        with pytest.raises(TypeError):
            ColorMap('test-name')

    def test_cmaps(self):
        count = 0
        for attribute_name in dir(ginga.cmap):
            if attribute_name.startswith('cmap_'):
                count = count + 1

        expected = count
        actual = len(ginga.cmap.cmaps)  # Can include matplotlib colormaps
        assert expected <= actual

    def test_add_cmap(self):
        test_clst = tuple([(x, x, x)
                           for x in np.linspace(0, 1, ginga.cmap.min_cmap_len)])
        ginga.cmap.add_cmap('test-name', test_clst)

        expected = ColorMap('test-name', test_clst)
        actual = ginga.cmap.cmaps['test-name']
        assert expected.name == actual.name
        assert expected.clst == actual.clst

        # Teardown
        del ginga.cmap.cmaps['test-name']

    def test_add_cmap_exception(self):
        test_clst = ((0.0, 0.0, 0.0), (1.0, 1.0, 1.0))
        with pytest.raises(AssertionError):
            ginga.cmap.add_cmap('test-name', test_clst)

    def test_get_cmap(self):
        test_clst = tuple([(x, x, x)
                           for x in np.linspace(0, 1, ginga.cmap.min_cmap_len)])
        ginga.cmap.add_cmap('test-name', test_clst)

        expected = ColorMap('test-name', test_clst)
        actual = ginga.cmap.get_cmap('test-name')
        assert expected.name == actual.name
        assert expected.clst == actual.clst

        # Teardown
        del ginga.cmap.cmaps['test-name']

    def test_get_cmap_exception(self):
        with pytest.raises(KeyError):
            ginga.cmap.get_cmap('non-existent-name')

    def test_get_names(self):
        names = []
        for attribute_name in dir(ginga.cmap):
            if attribute_name.startswith('cmap_'):
                names.append(attribute_name[5:])

        expected = set(names)
        actual = set(ginga.cmap.get_names())  # Can include matplotlib names
        assert expected <= actual

    # TODO: Add tests for matplotlib functions

# END



#
# log.py -- logging routines for Ginga
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import os
import logging
import logging.handlers

LOG_FORMAT = '%(asctime)s | %(levelname)1.1s | %(filename)s:%(lineno)d (%(funcName)s) | %(message)s'

# max size of log file before rotating
log_maxsize = 20 * 1024 * 1024
log_backups = 4


class NullLogger(object):
    """
    The NullLogger can be used in the place of a "real" logging module logger
    if the code just uses the standard levels/methods for logging.

    It is useful when you need to suppress logging or bypass the overhead of
    the logging module logger.
    """
    def __init__(self, f_out=None):
        self.f_out = f_out
        self.handlers = []

    def debug(self, msg):
        if self.f_out:
            self.f_out.write("| D | %s\n" % msg)
            self.f_out.flush()

    def info(self, msg):
        if self.f_out:
            self.f_out.write("| I | %s\n" % msg)
            self.f_out.flush()

    def warning(self, msg):
        if self.f_out:
            self.f_out.write("| W | %s\n" % msg)
            self.f_out.flush()

    def warn(self, msg):
        return self.warning(msg)

    def error(self, msg, exc_info=False):
        if self.f_out:
            self.f_out.write("| E | %s\n" % msg)
            self.f_out.flush()

    def critical(self, msg):
        if self.f_out:
            self.f_out.write("| C | %s\n" % msg)
            self.f_out.flush()

    def addHandler(self, hndlr):
        pass


def get_logger(name='ginga', level=None, null=False,
               options=None, log_file=None, log_stderr=False):

    if null or ((options is not None) and hasattr(options, 'nulllogger') and
                options.nulllogger):
        # User wants a Null Logger
        return NullLogger()

    # Create a logger
    logger = logging.Logger('ginga')

    if level is None:
        if (options is not None) and (options.loglevel is not None):
            level = options.loglevel
        else:
            level = logging.WARN

    fmt = logging.Formatter(LOG_FORMAT)
    if (not log_file) and (options is not None) and (options.logfile is not None):
        log_file = options.logfile

    if log_file is not None:
        if ((options is not None) and (getattr(options, 'rmlog', False)) and
                os.path.exists(log_file)):
            os.remove(log_file)
        # TODO: get maxsize and backup from options, if present
        fileHdlr = logging.handlers.RotatingFileHandler(
            log_file, maxBytes=log_maxsize, backupCount=log_backups)
        fileHdlr.setLevel(level)
        fileHdlr.setFormatter(fmt)
        logger.addHandler(fileHdlr)

    if (not log_stderr) and (options is not None) and (options.logstderr):
        log_stderr = options.logstderr

    if log_stderr:
        stderrHdlr = logging.StreamHandler()
        stderrHdlr.setLevel(level)
        stderrHdlr.setFormatter(fmt)
        logger.addHandler(stderrHdlr)

    return logger


def addlogopts(parser):
    if hasattr(parser, 'add_option'):
        # older OptParse
        add_argument = parser.add_option
    else:
        # newer ArgParse
        add_argument = parser.add_argument

    add_argument("--log", dest="logfile", metavar="FILE",
                 help="Write logging output to FILE")
    add_argument("--loglevel", dest="loglevel", metavar="LEVEL",
                 default=20, type=int,
                 help="Set logging level to LEVEL")
    add_argument("--lognull", dest="nulllogger", default=False,
                 action="store_true",
                 help="Use a null logger")
    add_argument("--logsize", dest="logsize", metavar="NUMBYTES",
                 type=int, default=log_maxsize,
                 help="Set maximum logging level to NUMBYTES")
    add_argument("--logbackups", dest="logbackups", metavar="NUM",
                 type=int, default=log_backups,
                 help="Set maximum number of backups to NUM")
    add_argument("--rmlog", dest="rmlog", default=False,
                 action="store_true",
                 help="Remove log if present (don't append)")
    add_argument("--stderr", dest="logstderr", default=False,
                 action="store_true",
                 help="Copy logging also to stderr")


#END



"""
Provides a base class for a 'dummy' setup.py command that has no functionality
(probably due to a missing requirement).  This dummy command can raise an
exception when it is run, explaining to the user what dependencies must be met
to use this command.

The reason this is at all tricky is that we want the command to be able to
provide this message even when the user passes arguments to the command.  If we
don't know ahead of time what arguments the command can take, this is
difficult, because distutils does not allow unknown arguments to be passed to a
setup.py command.  This hacks around that restriction to provide a useful error
message even when a user passes arguments to the dummy implementation of a
command.

Use this like:

    try:
        from some_dependency import SetupCommand
    except ImportError:
        from ._dummy import _DummyCommand

        class SetupCommand(_DummyCommand):
            description = \
                'Implementation of SetupCommand from some_dependency; '
                'some_dependency must be installed to run this command'

            # This is the message that will be raised when a user tries to
            # run this command--define it as a class attribute.
            error_msg = \
                "The 'setup_command' command requires the some_dependency "
                "package to be installed and importable."
"""

import sys
from setuptools import Command
from distutils.errors import DistutilsArgError
from textwrap import dedent


class _DummyCommandMeta(type):
    """
    Causes an exception to be raised on accessing attributes of a command class
    so that if ``./setup.py command_name`` is run with additional command-line
    options we can provide a useful error message instead of the default that
    tells users the options are unrecognized.
    """

    def __init__(cls, name, bases, members):
        if bases == (Command, object):
            # This is the _DummyCommand base class, presumably
            return

        if not hasattr(cls, 'description'):
            raise TypeError(
                "_DummyCommand subclass must have a 'description' "
                "attribute.")

        if not hasattr(cls, 'error_msg'):
            raise TypeError(
                "_DummyCommand subclass must have an 'error_msg' "
                "attribute.")

    def __getattribute__(cls, attr):
        if attr in ('description', 'error_msg'):
            # Allow cls.description to work so that `./setup.py
            # --help-commands` still works
            return super(_DummyCommandMeta, cls).__getattribute__(attr)

        raise DistutilsArgError(cls.error_msg)


class _DummyCommand(Command, object, metaclass=_DummyCommandMeta):
    pass



#
# Plot.py -- Plotting widget canvas wrapper.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
from io import BytesIO
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

from ginga.web.pgw import Widgets


class PlotWidget(Widgets.Canvas):
    """
    This class implements the server-side backend of the surface for a
    web-based plot viewer.  It uses a web socket to connect to an HTML5
    canvas with javascript callbacks in a web browser on the client.

    The viewer is created separately on the backend and connects to this
    surface via the set_viewer() method.
    """

    def __init__(self, plot, width=500, height=500):
        super(PlotWidget, self).__init__(width=width, height=height)

        self.widget = FigureCanvas(plot.get_figure())
        self.logger = plot.logger

        self._configured = False

        self.set_plot(plot)

    def set_plot(self, plot):
        self.plot = plot
        self.logger = plot.logger
        self.logger.debug("set_plot called")

        self._dispatch_event_table = {
            "activate": self.ignore_event,
            "setbounds": self.map_event_cb,
            "mousedown": self.ignore_event,
            "mouseup": self.ignore_event,
            "mousemove": self.ignore_event,
            "mouseout": self.ignore_event,
            "mouseover": self.ignore_event,
            "mousewheel": self.ignore_event,
            "wheel": self.ignore_event,
            "click": self.ignore_event,
            "dblclick": self.ignore_event,
            "keydown": self.ignore_event,
            "keyup": self.ignore_event,
            "keypress": self.ignore_event,
            "resize": self.resize_event,
            "focus": self.ignore_event,
            "focusout": self.ignore_event,
            "blur": self.ignore_event,
            "drop": self.ignore_event,
            "paste": self.ignore_event,
            # Hammer.js events
            "pinch": self.ignore_event,
            "pinchstart": self.ignore_event,
            "pinchend": self.ignore_event,
            "rotate": self.ignore_event,
            "rotatestart": self.ignore_event,
            "rotateend": self.ignore_event,
            "tap": self.ignore_event,
            "pan": self.ignore_event,
            "panstart": self.ignore_event,
            "panend": self.ignore_event,
            "swipe": self.ignore_event,
        }

        self.plot.add_callback('draw-canvas', self.draw_cb)

    def get_plot(self):
        return self.plot

    def ignore_event(self, event):
        pass

    def do_refresh(self):
        app = self.get_app()
        app.do_operation('refresh_canvas', id=self.id)

    def get_rgb_buffer(self, plot):
        buf = BytesIO()
        fig = plot.get_figure()
        fig.canvas.print_figure(buf, format='png')
        wd, ht = self.width, self.height
        return (wd, ht, buf.getvalue())

    def draw_cb(self, plot):
        self.logger.debug("getting RGB buffer")
        wd, ht, buf = self.get_rgb_buffer(plot)

        #self.logger.debug("clear_rect")
        #self.clear_rect(0, 0, wd, ht)

        self.logger.debug("drawing %dx%d image" % (wd, ht))
        self.draw_image(buf, 0, 0, wd, ht)

    def configure_window(self, wd, ht):
        self.logger.debug("canvas resized to %dx%d" % (wd, ht))
        fig = self.plot.get_figure()
        fig.set_size_inches(float(wd) / fig.dpi, float(ht) / fig.dpi)

    def map_event_cb(self, event):
        wd, ht = event.width, event.height
        self.configure_window(wd, ht)

        self.plot.draw()
        self.do_refresh()

    def resize_event(self, event):
        wd, ht = event.width, event.height
        self.configure_window(wd, ht)

        self.plot.draw()

    def _cb_redirect(self, event):
        method = self._dispatch_event_table[event.type]
        try:
            method(event)

        except Exception as e:
            self.logger.error("error redirecting '%s' event: %s" % (
                event.type, str(e)))
            # TODO: dump traceback to debug log


#END



"""Download rendered HTML doc from RTD."""
import os
import shutil
import zipfile
import urllib

from astropy.utils import minversion
from astropy.utils.data import _find_pkg_data_path

from ginga import toolkit

__all__ = ['get_doc']


def _find_rtd_version():
    """Find closest RTD doc version."""
    vstr = 'latest'
    try:
        import ginga
        from bs4 import BeautifulSoup
    except ImportError:
        return vstr

    # No active doc build before this release, just use latest.
    if not minversion(ginga, '2.6.0'):
        return vstr

    # Get RTD download listing.
    url = 'https://readthedocs.org/projects/ginga/downloads/'
    with urllib.request.urlopen(url) as r:
        soup = BeautifulSoup(r, 'html.parser')

    # Compile a list of available HTML doc versions for download.
    all_rtd_vernums = []
    for link in soup.find_all('a'):
        href = link.get('href')
        if 'htmlzip' not in href:
            continue
        s = href.split('/')[-2]
        if s.startswith('v'):  # Ignore latest and stable
            all_rtd_vernums.append(s)
    all_rtd_vernums.sort(reverse=True)

    # Find closest match.
    ginga_ver = ginga.__version__
    for rtd_ver in all_rtd_vernums:
        if ginga_ver > rtd_ver[1:]:  # Ignore "v" in comparison
            break
        else:
            vstr = rtd_ver

    return vstr


def _download_rtd_zip(rtd_version=None, **kwargs):
    """
    Download and extract HTML ZIP from RTD to installed doc data path.
    Download is skipped if content already exists.

    Parameters
    ----------
    rtd_version : str or `None`
        RTD version to download; e.g., "latest", "stable", or "v2.6.0".
        If not given, download closest match to software version.

    kwargs : dict
        Keywords for ``urlretrieve()``.

    Returns
    -------
    index_html : str
        Path to local "index.html".

    """
    # https://github.com/ejeschke/ginga/pull/451#issuecomment-298403134
    if not toolkit.family.startswith('qt'):
        raise ValueError('Downloaded documentation not compatible with {} '
                         'UI toolkit browser'.format(toolkit.family))

    if rtd_version is None:
        rtd_version = _find_rtd_version()

    data_path = os.path.dirname(
        _find_pkg_data_path('help.html', package='ginga.doc'))
    index_html = os.path.join(data_path, 'index.html')

    # There is a previous download of documentation; Do nothing.
    # There is no check if downloaded version is outdated; The idea is that
    # this folder would be empty again when installing new version.
    if os.path.isfile(index_html):
        return index_html

    url = ('https://readthedocs.org/projects/ginga/downloads/htmlzip/'
           '{}/'.format(rtd_version))
    local_path = urllib.request.urlretrieve(url, **kwargs)[0]

    with zipfile.ZipFile(local_path, 'r') as zf:
        zf.extractall(data_path)

    # RTD makes an undesirable sub-directory, so move everything there
    # up one level and delete it.
    subdir = os.path.join(data_path, 'ginga-{}'.format(rtd_version))
    for s in os.listdir(subdir):
        src = os.path.join(subdir, s)
        if os.path.isfile(src):
            shutil.copy(src, data_path)
        else:  # directory
            shutil.copytree(src, os.path.join(data_path, s))
    shutil.rmtree(subdir)

    if not os.path.isfile(index_html):
        raise OSError(
            '{} is missing; Ginga doc download failed'.format(index_html))

    return index_html


def get_doc(logger=None, plugin=None, reporthook=None):
    """
    Return URL to documentation. Attempt download if does not exist.

    Parameters
    ----------
    logger : obj or `None`
        Ginga logger.

    plugin : obj or `None`
        Plugin object. If given, URL points to plugin doc directly.
        If this function is called from within plugin class,
        pass ``self`` here.

    reporthook : callable or `None`
        Report hook for ``urlretrieve()``.

    Returns
    -------
    url : str or `None`
        URL to local documentation, if available.

    """
    from ginga.GingaPlugin import GlobalPlugin, LocalPlugin

    if isinstance(plugin, GlobalPlugin):
        plugin_page = 'plugins_global'
        plugin_name = str(plugin)
    elif isinstance(plugin, LocalPlugin):
        plugin_page = 'plugins_local'
        plugin_name = str(plugin)
    else:
        plugin_page = None
        plugin_name = None

    try:
        index_html = _download_rtd_zip(reporthook=reporthook)

    # Download failed, use online resource
    except Exception as e:
        url = 'https://ginga.readthedocs.io/en/latest/'

        if plugin_name is not None:
            if toolkit.family.startswith('qt'):
                # This displays plugin docstring.
                url = None
            else:
                # This redirects to online doc.
                url += 'manual/{}/{}.html'.format(plugin_page, plugin_name)

        if logger is not None:
            logger.error(str(e))

    # Use local resource
    else:
        pfx = 'file:'
        url = '{}{}'.format(pfx, index_html)

        # https://github.com/rtfd/readthedocs.org/issues/2803
        if plugin_name is not None:
            url += '#{}'.format(plugin_name)

    return url



#
# Plot.py -- Plotting function for Ginga viewer.
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import matplotlib
# GTK3Cairo backend is slow to redraw compared to GTK3Agg!
# matplotlib.use('GTK3Cairo')
# from matplotlib.backends.backend_gtk3cairo import (FigureCanvasGTK3Cairo
#                                                    as FigureCanvas)  # noqa
matplotlib.use('GTK3Agg')
from matplotlib.backends.backend_gtk3agg import (FigureCanvasGTK3Agg
                                                 as FigureCanvas)  # noqa
from ginga.gtk3w import Widgets  # noqa


class PlotWidget(Widgets.WidgetBase):

    def __init__(self, plot, width=500, height=500):
        super(PlotWidget, self).__init__()

        self.widget = FigureCanvas(plot.get_figure())
        self.plot = plot
        self.logger = plot.logger

        self.widget.set_size_request(width, height)
        self.widget.show_all()

    def set_plot(self, plot):
        self.plot = plot
        self.logger = plot.logger
        self.logger.debug("set_plot called")

    def configure_window(self, wd, ht):
        self.logger.debug("canvas resized to %dx%d" % (wd, ht))
        fig = self.plot.get_figure()
        fig.set_size_inches(float(wd) / fig.dpi, float(ht) / fig.dpi)

# END



#
# RGBMap.py -- color mapping
#
# This is open-source software licensed under a BSD license.
# Please see the file LICENSE.txt for details.
#
import numpy as np

from ginga.misc import Callback, Settings
from ginga import ColorDist, trcalc
from ginga import cmap as mod_cmap
from ginga import imap as mod_imap


class RGBMapError(Exception):
    pass


class RGBPlanes(object):

    def __init__(self, rgbarr, order):
        self.rgbarr = rgbarr
        order = order.upper()
        self.order = order
        self.hasAlpha = 'A' in order

    def get_slice(self, ch):
        return self.rgbarr[..., self.order.index(ch.upper())]

    def has_slice(self, ch):
        return ch.upper() in self.order

    def get_order(self):
        return self.order

    def get_order_indexes(self, cs):
        cs = cs.upper()
        return [self.order.index(c) for c in cs]

    def get_array(self, order, dtype=None):
        """Get Numpy array that represents the RGB layers.

        Parameters
        ----------
        order : str
            The desired order of RGB color layers.

        Returns
        -------
        arr : ndarray
            Numpy array that represents the RGB layers.

        dtype : numpy array dtype (optional)
            Type of the numpy array desired.

        """
        if dtype is None:
            dtype = self.rgbarr.dtype
        order = order.upper()
        if order == self.order:
            return self.rgbarr.astype(dtype, copy=False)

        res = trcalc.reorder_image(order, self.rgbarr, self.order)
        res = res.astype(dtype, copy=False, casting='unsafe')
        return res

    def get_size(self):
        """Returns (height, width) tuple of slice size."""
        return self.get_slice('R').shape


class RGBMapper(Callback.Callbacks):

    ############################################################
    # CODE NOTES
    #
    # [A] Some numpy routines have been optomized by using the out=
    # parameter, which avoids having to allocate a new array for the
    # result
    #

    def __init__(self, logger, dist=None, settings=None, bpp=None):
        Callback.Callbacks.__init__(self)

        self.logger = logger

        # Create settings and set defaults
        if settings is None:
            settings = Settings.SettingGroup(logger=self.logger)
        self.settings = settings
        self.t_ = settings
        self.settings_keys = ['color_map', 'intensity_map',
                              'color_array', 'shift_array',
                              'color_algorithm', 'color_hashsize',
                              ]

        # add our defaults
        self.t_.add_defaults(color_map='gray', intensity_map='ramp',
                             color_algorithm='linear',
                             color_hashsize=65535,
                             color_array=None, shift_array=None)
        self.t_.get_setting('color_map').add_callback('set',
                                                      self.color_map_set_cb)
        self.t_.get_setting('intensity_map').add_callback('set',
                                                          self.intensity_map_set_cb)
        self.t_.get_setting('color_array').add_callback('set',
                                                        self.color_array_set_cb)
        self.t_.get_setting('shift_array').add_callback('set',
                                                        self.shift_array_set_cb)
        self.t_.get_setting('color_hashsize').add_callback('set',
                                                           self.color_hashsize_set_cb)
        self.t_.get_setting('color_algorithm').add_callback('set',
                                                            self.color_algorithm_set_cb)

        # For color and intensity maps
        self.cmap = None
        self.imap = None
        self.arr = None
        self.iarr = None
        self.carr = None
        self.sarr = None
        self.scale_pct = 1.0

        # targeted bit depth per-pixel band of the output RGB array
        # (can be less than the data size of the output array)
        if bpp is None:
            bpp = 8
        self.bpp = bpp
        # maximum value that we can generate in this range
        self.maxc = int(2 ** self.bpp - 1)
        # data size per pixel band in the output RGB array
        self._set_dtype()

        # For scaling algorithms
        hashsize = self.t_.get('color_hashsize', 65536)
        if dist is None:
            color_alg_name = self.t_.get('color_algorithm', 'linear')
            color_dist_class = ColorDist.get_dist(color_alg_name)
            dist = color_dist_class(hashsize)
        self.dist = dist

        # For callbacks
        for name in ('changed', ):
            self.enable_callback(name)

        self.suppress_changed = self.suppress_callback('changed', 'last')

        carr = self.t_.get('color_array', None)
        sarr = self.t_.get('shift_array', None)

        cm_name = self.t_.get('color_map', 'gray')
        self.set_color_map(cm_name)

        im_name = self.t_.get('intensity_map', 'ramp')
        self.set_intensity_map(im_name)

        # Initialize color array
        if carr is not None:
            self.set_carr(carr)
        else:
            self.calc_cmap()

        # Initialize shift array
        if sarr is not None:
            self.set_sarr(sarr)
        else:
            self.reset_sarr(callback=False)

    def _set_dtype(self):
        if self.bpp <= 8:
            self.dtype = np.uint8
        elif self.bpp <= 16:
            self.dtype = np.uint16
        else:
            self.dtype = np.uint32

    def set_bpp(self, bpp):
        """
        Set the bit depth (per band) for the output.
        A typical "32-bit RGBA" image would be 8; a 48-bit image would
        be 16, etc.
        """
        self.bpp = bpp
        self.maxc = int(2 ** self.bpp - 1)
        self._set_dtype()
        self.calc_cmap()
        self.recalc(callback=False)

    def get_settings(self):
        return self.t_

    def set_color_map(self, cmap_name):
        self.t_.set(color_map=cmap_name)

    def color_map_set_cb(self, setting, cmap_name):
        cm = mod_cmap.get_cmap(cmap_name)
        self.set_cmap(cm, callback=True)

    def set_cmap(self, cmap, callback=True):
        """
        Set the color map used by this RGBMapper.

        `cmap` specifies a ColorMap object.  If `callback` is True, then
        any callbacks associated with this change will be invoked.
        """
        self.cmap = cmap
        with self.suppress_changed:
            self.calc_cmap()
            # TEMP: ignore passed callback parameter
            # callback=False in the following because we don't want to
            # recursively invoke set_cmap()
            self.t_.set(color_map=cmap.name, callback=False)

    def get_cmap(self):
        """
        Return the color map used by this RGBMapper.
        """
        return self.cmap

    def calc_cmap(self):
        clst = self.cmap.clst
        self.maxc = len(clst) - 1
        arr = np.array(clst).transpose() * float(self.maxc)
        # does this really need to be the same type as rgbmap output type?
        carr = np.round(arr).astype(self.dtype, copy=False)
        self.t_.set(color_array=carr)

    def invert_cmap(self, callback=True):
        carr = np.fliplr(self.carr)
        # TEMP: ignore passed callback parameter
        self.t_.set(color_array=carr)

    def rotate_cmap(self, num, callback=True):
        carr = np.roll(self.carr, num, axis=1)
        # TEMP: ignore passed callback parameter
        self.t_.set(color_array=carr)

    def restore_cmap(self, callback=True):
        with self.suppress_changed:
            self.reset_sarr(callback=False)
            # TEMP: ignore passed callback parameter
            self.calc_cmap()

    def reset_cmap(self):
        self.recalc()

    def get_rgb(self, index):
        """
        Return a tuple of (R, G, B) values in the 0-maxc range associated
        mapped by the value of `index`.
        """
        index = int(index)
        return tuple(self.arr[index])

    def get_rgbval(self, index):
        """
        Return a tuple of (R, G, B) values in the 0-maxc range associated
        mapped by the value of `index`.
        """
        index = int(index)
        assert (index >= 0) and (index <= self.maxc), \
            RGBMapError("Index must be in range 0-%d !" % (self.maxc))
        index = int(self.sarr[index].clip(0, self.maxc))
        return (self.arr[0][index],
                self.arr[1][index],
                self.arr[2][index])

    def set_intensity_map(self, imap_name):
        self.t_.set(intensity_map=imap_name)

    def intensity_map_set_cb(self, setting, imap_name):
        im = mod_imap.get_imap(imap_name)
        self.set_imap(im, callback=True)

    def set_imap(self, imap, callback=True):
        """
        Set the intensity map used by this RGBMapper.

        `imap` specifies an IntensityMap object.  If `callback` is True, then
        any callbacks associated with this change will be invoked.
        """
        self.imap = imap
        self.calc_imap()
        with self.suppress_changed:
            # TEMP: ignore passed callback parameter
            self.recalc()
            # callback=False in the following because we don't want to
            # recursively invoke set_imap()
            self.t_.set(intensity_map=imap.name, callback=False)

    def get_imap(self):
        """
        Return the intensity map used by this RGBMapper.
        """
        return self.imap

    def calc_imap(self):
        arr = np.array(self.imap.ilst) * float(self.maxc)
        self.iarr = np.round(arr).astype(np.uint, copy=False)

    def reset_sarr(self, callback=True):
        maxlen = self.maxc + 1
        self.scale_pct = 1.0
        sarr = np.arange(maxlen)
        # TEMP: ignore passed callback parameter
        self.t_.set(shift_array=sarr)

    def get_sarr(self):
        return self.sarr

    def set_sarr(self, sarr, callback=True):
        if sarr is not None:
            sarr = np.asarray(sarr, dtype=np.uint)
        # TEMP: ignore passed callback parameter
        self.t_.set(shift_array=sarr)

    def shift_array_set_cb(self, setting, sarr):
        if sarr is not None:
            sarr = np.asarray(sarr)
            maxlen = self.maxc + 1
            _len = len(sarr)
            if _len != maxlen:
                raise RGBMapError("shift map length %d != %d" % (_len, maxlen))
            self.sarr = sarr.astype(np.uint, copy=False)
            # NOTE: can't reset scale_pct here because it results in a
            # loop with e.g. scale_and_shift()
            #self.scale_pct = 1.0
            self.make_callback('changed')
        else:
            self.reset_sarr(callback=True)

    def get_carr(self):
        return self.carr

    def set_carr(self, carr, callback=True):
        if carr is not None:
            carr = np.asarray(carr)
        # TEMP: ignore passed callback parameter
        self.t_.set(color_array=carr, callback=True)

    def color_array_set_cb(self, setting, carr):
        if carr is not None:
            carr = np.asarray(carr)
            maxlen = self.maxc + 1
            self.carr = carr.astype(self.dtype, copy=False)
            _len = carr.shape[1]
            if _len != maxlen:
                raise RGBMapError("color map length %d != %d" % (_len, maxlen))
        else:
            self.calc_cmap()
        self.recalc(callback=True)

    def recalc(self, callback=True):
        self.arr = np.copy(self.carr)
        # Apply intensity map to rearrange colors
        if self.iarr is not None:
            idx = self.iarr
            self.arr[0] = self.arr[0][idx]
            self.arr[1] = self.arr[1][idx]
            self.arr[2] = self.arr[2][idx]

        # NOTE: don't reset shift array
        #self.reset_sarr(callback=False)
        if callback:
            self.make_callback('changed')

    def get_hash_size(self):
        return self.dist.get_hash_size()

    def set_hash_size(self, size, callback=True):
        # TEMP: ignore passed callback parameter
        self.t_.set(color_hashsize=size)

    def color_hashsize_set_cb(self, setting, size):
        self.dist.set_hash_size(size)
        self.make_callback('changed')

    def get_hash_algorithms(self):
        return ColorDist.get_dist_names()

    def get_hash_algorithm(self):
        return str(self.dist)

    def get_dist(self):
        """
        Return the color distribution used by this RGBMapper.
        """
        return self.dist

    def set_dist(self, dist, callback=True):
        self.dist = dist
        if callback:
            self.make_callback('changed')

    def set_hash_algorithm(self, name, callback=True, **kwargs):
        # TODO: deal with algorithm parameters in kwargs
        self.t_.set(color_algorithm=name)

    def color_algorithm_set_cb(self, setting, name):
        size = self.t_.get('color_hashsize', self.dist.get_hash_size())
        dist = ColorDist.get_dist(name)(size)
        self.set_dist(dist, callback=True)

    def get_order_indexes(self, order, cs):
        order = order.upper()
        if order == '':
            # assume standard RGB order if we don't find an order
            # explicitly set
            return [0, 1, 2]
        cs = cs.upper()
        return [order.index(c) for c in cs]

    def _get_rgbarray(self, idx, rgbobj, image_order=''):
        # NOTE: data is assumed to be in the range 0-maxc at this point
        # but clip as a precaution
        # See NOTE [A]: idx is always an array calculated in the caller and
        #    discarded afterwards
        # idx = idx.clip(0, self.maxc).astype(np.uint, copy=False)
        idx.clip(0, self.maxc, out=idx)

        # run it through the shift array and clip the result
        # See NOTE [A]
        # idx = self.sarr[idx].clip(0, self.maxc).astype(np.uint, copy=False)
        idx = self.sarr[idx]
        idx.clip(0, self.maxc, out=idx)

        ri, gi, bi = self.get_order_indexes(rgbobj.get_order(), 'RGB')
        out = rgbobj.rgbarr

        # change [A]
        if (image_order is None) or (len(image_order) < 3):
            out[..., ri] = self.arr[0][idx]
            out[..., gi] = self.arr[1][idx]
            out[..., bi] = self.arr[2][idx]
        else:
            # <== indexes already contain RGB info.
            rj, gj, bj = self.get_order_indexes(image_order, 'RGB')
            out[..., ri] = self.arr[0][idx[..., rj]]
            out[..., gi] = self.arr[1][idx[..., gj]]
            out[..., bi] = self.arr[2][idx[..., bj]]

    def get_rgbarray(self, idx, out=None, order='RGB', image_order=''):
        """
        Parameters
        ----------
        idx : index array

        out : output array or None
            The output array.  If `None` one of the correct size and depth
            will be created.

        order : str
            The order of the color planes in the output array (e.g. "ARGB")

        image_order : str or None
            The order of channels if indexes already contain RGB info.
        """
        # prepare output array
        shape = idx.shape
        depth = len(order)

        if (image_order is not None) and (len(image_order) > 1):
            # indexes contain RGB axis, so omit this
            res_shape = shape[:-1] + (depth, )
        else:
            res_shape = shape + (depth, )

        if out is None:
            out = np.empty(res_shape, dtype=self.dtype, order='C')
        else:
            # TODO: assertion check on shape of out
            assert res_shape == out.shape, \
                RGBMapError("Output array shape %s doesn't match result "
                            "shape %s" % (str(out.shape), str(res_shape)))

        res = RGBPlanes(out, order)

        # set alpha channel
        if res.hasAlpha:
            aa = res.get_slice('A')
            aa.fill(self.maxc)

        idx = self.get_hasharray(idx)

        self._get_rgbarray(idx, res, image_order=image_order)

        return res

    def get_hasharray(self, idx):
        return self.dist.hash_array(idx)

    def _shift(self, sarr, pct, rotate=False):
        n = len(sarr)
        num = int(n * pct)
        arr = np.roll(sarr, num)
        if not rotate:
            if num > 0:
                arr[0:num] = sarr[0]
            elif num < 0:
                arr[n + num:n] = sarr[-1]
        return arr

    def _stretch(self, sarr, scale):
        old_wd = len(sarr)
        new_wd = int(round(scale * old_wd))

        # Is there a more efficient way to do this?
        xi = np.mgrid[0:new_wd]
        iscale_x = float(old_wd) / float(new_wd)

        xi = (xi * iscale_x).astype(np.uint, copy=False)
        xi = xi.clip(0, old_wd - 1).astype(np.uint, copy=False)
        newdata = sarr[xi]
        return newdata

    def shift(self, pct, rotate=False, callback=True):
        work = self._shift(self.sarr, pct, rotate=rotate)
        maxlen = self.maxc + 1
        assert len(work) == maxlen, \
            RGBMapError("shifted shift map is != %d" % maxlen)
        self.t_.set(shift_array=work)

    def scale_and_shift(self, scale_pct, shift_pct, callback=True):
        """Stretch and/or shrink the color map via altering the shift map.
        """
        maxlen = self.maxc + 1
        self.sarr = np.arange(maxlen)

        # limit shrinkage to 5% of original size
        scale = max(scale_pct, 0.050)
        self.scale_pct = scale

        work = self._stretch(self.sarr, scale)
        n = len(work)
        if n < maxlen:
            # pad on the lowest and highest values of the shift map
            m = (maxlen - n) // 2 + 1
            barr = np.array([0] * m)
            tarr = np.array([self.maxc] * m)
            work = np.concatenate([barr, work, tarr])
            work = work[:maxlen]

        # we are mimicing ds9's stretch and shift algorithm here.
        # ds9 seems to cut the center out of the stretched array
        # BEFORE shifting
        n = len(work) // 2
        halflen = maxlen // 2
        work = work[n - halflen:n + halflen].astype(np.uint, copy=False)
        assert len(work) == maxlen, \
            RGBMapError("scaled shift map is != %d" % maxlen)

        # shift map according to the shift_pct
        work = self._shift(work, shift_pct)
        assert len(work) == maxlen, \
            RGBMapError("shifted shift map is != %d" % maxlen)

        self.t_.set(shift_array=work)

    def stretch(self, scale_factor, callback=True):
        """Stretch the color map via altering the shift map.
        """
        self.scale_pct *= scale_factor
        self.scale_and_shift(self.scale_pct, 0.0, callback=callback)

    def copy_attributes(self, dst_rgbmap):
        self.t_.copy_settings(dst_rgbmap.get_settings(),
                              keylist=self.settings_keys)

    def share_attributes(self, dst_rgbmap):
        self.t_.share_settings(dst_rgbmap.get_settings(),
                               keylist=self.settings_keys)


class NonColorMapper(RGBMapper):
    """
    A kind of color mapper for data that is already in RGB form.

    This mapper allows changing of color distribution and contrast
    adjustment, but does no coloring.
    """
    def __init__(self, logger, dist=None, bpp=None):
        super(NonColorMapper, self).__init__(logger, dist=dist, bpp=bpp)

        maxlen = self.maxc + 1
        self.dist.set_hash_size(maxlen)
        self.reset_sarr(callback=False)

    def _get_rgbarray(self, idx, rgbobj, image_order='RGB'):
        # NOTE: data is assumed to be in the range 0-maxc at this point
        # but clip as a precaution
        # See NOTE [A]: idx is always an array calculated in the caller and
        #    discarded afterwards
        idx.clip(0, self.maxc, out=idx)

        # run it through the shift array and clip the result
        # See NOTE [A]
        idx = self.sarr[idx]
        idx.clip(0, self.maxc, out=idx)

        ri, gi, bi = self.get_order_indexes(rgbobj.get_order(), 'RGB')
        rj, gj, bj = self.get_order_indexes(image_order, 'RGB')
        out = rgbobj.rgbarr

        out[..., ri] = idx[..., rj]
        out[..., gi] = idx[..., gj]
        out[..., bi] = idx[..., bj]


class PassThruRGBMapper(RGBMapper):
    """
    A kind of color mapper for data that is already in RGB form.

    This mapper bypasses color distribution, contrast adjustment and
    coloring.  It is thus the most efficient one to use for maximum
    speed rendering of "finished" RGB data.
    """
    def __init__(self, logger, dist=None, bpp=None):
        super(PassThruRGBMapper, self).__init__(logger, bpp=bpp)

        # ignore passed in distribution
        maxlen = self.maxc + 1
        self.dist = ColorDist.LinearDist(maxlen)

    def get_hasharray(self, idx):
        # data is already constrained to 0..maxc and we want to
        # bypass color redistribution
        return idx

    def _get_rgbarray(self, idx, rgbobj, image_order='RGB'):
        # NOTE: data is assumed to be in the range 0-maxc at this point
        # but clip as a precaution
        # See NOTE [A]: idx is always an array calculated in the caller and
        #    discarded afterwards
        idx.clip(0, self.maxc, out=idx)

        # bypass the shift array and skip color mapping,
        # index is the final data
        ri, gi, bi = self.get_order_indexes(rgbobj.get_order(), 'RGB')
        rj, gj, bj = self.get_order_indexes(image_order, 'RGB')
        out = rgbobj.rgbarr

        out[..., ri] = idx[..., rj]
        out[..., gi] = idx[..., gj]
        out[..., bi] = idx[..., bj]


#END
